from __future__ import annotations

import functools
from typing import Callable, Hashable, Mapping, Optional, TypeVar, Union

T = TypeVar("T")

AtomicObservation = Union[T, Callable[..., T]]  # TODO add support for more atomic types
CompoundObservation = Union[
    Mapping[Hashable, AtomicObservation[T]], Callable[..., AtomicObservation[T]]
]
Observation = Union[AtomicObservation[T], CompoundObservation[T]]


@functools.singledispatch
def observe(rv, obs: Optional[Observation[T]] = None, **kwargs) -> T:
    """
    Observe a random value in a probabilistic program.
    """
    raise NotImplementedError(f"observe not implemented for type {type(rv)}")



from __future__ import annotations

import functools
from typing import Mapping, Optional, TypeVar

import pyro
import pyro.distributions
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin
from typing_extensions import ParamSpec

from chirho.indexed.handlers import add_indices
from chirho.indexed.ops import IndexSet, get_index_plates, indices_of
from chirho.observational.ops import AtomicObservation, observe

P = ParamSpec("P")
K = TypeVar("K")
T = TypeVar("T")


@observe.register(int)
@observe.register(float)
@observe.register(bool)
@observe.register(torch.Tensor)
def _observe_deterministic(rv: T, obs: Optional[AtomicObservation[T]] = None, **kwargs):
    """
    Observe a tensor in a probabilistic program.
    """
    rv_dist = pyro.distributions.Delta(
        torch.as_tensor(rv), event_dim=kwargs.pop("event_dim", 0)
    )
    return observe(rv_dist, obs, **kwargs)


@observe.register(pyro.distributions.Distribution)
@pyro.poutine.runtime.effectful(type="observe")
def _observe_distribution(
    rv: TorchDistributionMixin,
    obs: Optional[AtomicObservation[torch.Tensor]] = None,
    *,
    name: Optional[str] = None,
    **kwargs,
) -> torch.Tensor:
    if name is None:
        raise ValueError("name must be specified when observing a distribution")

    if callable(obs):
        raise NotImplementedError("Dependent observations are not yet supported")

    return pyro.sample(name, rv, obs=obs, **kwargs)


@observe.register(dict)
def _observe_dict(
    rv: Mapping[K, T],
    obs: Optional[AtomicObservation[Mapping[K, T]]] = None,
    *,
    name: Optional[str] = None,
    **kwargs,
) -> Mapping[K, T]:
    if callable(obs):
        obs = obs(rv)
        if obs is not rv and obs is not None:
            raise NotImplementedError("Dependent observations are not yet supported")

    if obs is rv or obs is None:
        return rv

    return {k: observe(rv[k], obs[k], name=f"{name}__{k}", **kwargs) for k in rv.keys()}


class ObserveNameMessenger(pyro.poutine.messenger.Messenger):
    def _pyro_observe(self, msg):
        if "name" not in msg["kwargs"]:
            msg["kwargs"]["name"] = msg["name"]


def site_is_delta(msg: pyro.poutine.messenger.Message) -> bool:
    d = msg["fn"]
    while hasattr(d, "base_dist"):
        d = d.base_dist
    return isinstance(d, pyro.distributions.Delta)


@functools.singledispatch
def unbind_leftmost_dim(v, name: str, size: int = 1, **kwargs):
    """
    Helper function to move the leftmost dimension of a ``torch.Tensor``
    or ``pyro.distributions.Distribution`` or other batched value
    into a fresh named dimension using the machinery in ``chirho.indexed`` ,
    allocating a new dimension with the given name if necessary
    via an enclosing :class:`~chirho.indexed.handlers.IndexPlatesMessenger` .

    .. warning:: Must be used in conjunction with :class:`~chirho.indexed.handlers.IndexPlatesMessenger` .

    :param v: Batched value.
    :param name: Name of the fresh dimension.
    :param size: Size of the fresh dimension. If 1, the size is inferred from ``v`` .
    """
    raise NotImplementedError


@unbind_leftmost_dim.register
def _unbind_leftmost_dim_tensor(
    v: torch.Tensor, name: str, size: int = 1, *, event_dim: int = 0
) -> torch.Tensor:
    size = max(size, v.shape[0])
    v = v.expand((size,) + v.shape[1:])

    if name not in get_index_plates():
        add_indices(IndexSet(**{name: set(range(size))}))

    new_dim = get_index_plates()[name].dim
    assert new_dim is not None
    orig_shape = v.shape
    while new_dim - event_dim < -len(v.shape):
        v = v[None]
    if v.shape[0] == 1 and orig_shape[0] != 1:
        v = torch.transpose(v, -len(orig_shape), new_dim - event_dim)
    return v


@unbind_leftmost_dim.register
def _unbind_leftmost_dim_distribution(
    v: TorchDistributionMixin, name: str, size: int = 1, **kwargs
) -> TorchDistributionMixin:
    size = max(size, v.batch_shape[0])
    if v.batch_shape[0] != 1:
        raise NotImplementedError("Cannot freely reshape distribution")

    if name not in get_index_plates():
        add_indices(IndexSet(**{name: set(range(size))}))

    new_dim = get_index_plates()[name].dim
    assert new_dim is not None
    orig_shape = v.batch_shape

    new_shape = (size,) + (1,) * (-new_dim - len(orig_shape)) + orig_shape[1:]
    return v.expand(new_shape)


@functools.singledispatch
def bind_leftmost_dim(v, name: str, **kwargs):
    """
    Helper function to move a named dimension managed by ``chirho.indexed``
    into a new unnamed dimension to the left of all named dimensions in the value.

    .. warning:: Must be used in conjunction with :class:`~chirho.indexed.handlers.IndexPlatesMessenger` .
    """
    raise NotImplementedError


@bind_leftmost_dim.register
def _bind_leftmost_dim_tensor(
    v: torch.Tensor, name: str, *, event_dim: int = 0, **kwargs
) -> torch.Tensor:
    if name not in indices_of(v, event_dim=event_dim):
        return v

    plate_dim = get_index_plates()[name].dim
    assert plate_dim is not None
    return torch.transpose(v[None], -len(v.shape) - 1, plate_dim - event_dim)



from __future__ import annotations

from typing import Callable, Generic, Mapping, TypeVar, Union

import pyro
import torch

from chirho.observational.internals import ObserveNameMessenger
from chirho.observational.ops import Observation, observe

T = TypeVar("T")
R = Union[float, torch.Tensor]


class Factors(Generic[T], pyro.poutine.messenger.Messenger):
    """
    Effect handler that adds new log-factors to the unnormalized
    joint log-density of a probabilistic program.

    After a :func:`pyro.sample` site whose name appears in ``factors``,
    this handler inserts a new :func:`pyro.factor` site
    whose name is prefixed with the string ``prefix``
    and whose log-weight is the result of applying the corresponding function
    to the value of the sample site. ::

        >>> with Factors(factors={"x": lambda x: -(x - 1) ** 2}, prefix="__factor_"):
        ...   with pyro.poutine.trace() as tr:
        ...     x = pyro.sample("x", dist.Normal(0, 1))
        ... tr.trace.compute_log_prob()
        >>> assert {"x", "__factor_x"} <= set(tr.trace.nodes.keys())
        >>> assert torch.all(tr.trace.nodes["x"]["log_prob"] == -(x - 1) ** 2)

    :param factors: A mapping from sample site names to log-factor functions.
    :param prefix: The prefix to use for the names of the factor sites.
    """

    factors: Mapping[str, Callable[[T], R]]
    prefix: str

    def __init__(
        self,
        factors: Mapping[str, Callable[[T], R]],
        *,
        prefix: str = "__factor_",
    ):
        self.factors = factors
        self.prefix = prefix
        super().__init__()

    def _pyro_post_sample(self, msg: dict) -> None:
        try:
            factor = self.factors[msg["name"]]
        except KeyError:
            return

        pyro.factor(f"{self.prefix}{msg['name']}", torch.tensor(factor(msg["value"])))


class Observations(Generic[T], ObserveNameMessenger):
    """
    Condition on values in a probabilistic program.

    Can be used as a drop-in replacement for :func:`pyro.condition` that supports
    a richer set of observational data types and enables counterfactual inference.
    """

    data: Mapping[str, Observation[T]]

    def __init__(self, data: Mapping[str, Observation[T]]):
        self.data = data
        super().__init__()

    def _pyro_sample(self, msg):
        if pyro.poutine.util.site_is_subsample(msg) or pyro.poutine.util.site_is_factor(
            msg
        ):
            return

        if msg["name"] not in self.data or msg["infer"].get("_do_not_observe", None):
            if (
                "_markov_scope" in msg["infer"]
                and getattr(self, "_current_site", None) is not None
            ):
                msg["infer"]["_markov_scope"].pop(self._current_site, None)
            return

        msg["stop"] = True
        msg["done"] = True

        # flags to guarantee commutativity of condition, intervene, trace
        msg["mask"] = False
        msg["is_observed"] = False
        msg["infer"]["is_auxiliary"] = True
        msg["infer"]["_do_not_trace"] = True
        msg["infer"]["_do_not_intervene"] = True
        msg["infer"]["_do_not_observe"] = True

        with pyro.poutine.infer_config(
            config_fn=lambda msg_: {
                "_do_not_observe": msg["name"] == msg_["name"]
                or msg_["infer"].get("_do_not_observe", False)
            }
        ):
            try:
                self._current_site = msg["name"]
                msg["value"] = observe(
                    msg["fn"], self.data[msg["name"]], name=msg["name"], **msg["kwargs"]
                )
            finally:
                self._current_site = None


if isinstance(pyro.poutine.handlers._make_handler(Observations), tuple):
    # backwards compatibility
    condition = pyro.poutine.handlers._make_handler(Observations)[1]
else:

    @pyro.poutine.handlers._make_handler(Observations)
    def condition(fn: Callable, data: Mapping[str, Observation[T]]): ...



from typing import Generic, TypeVar

import pyro
import torch

from chirho.dynamical.internals._utils import (
    _squeeze_time_dim,
    _unsqueeze_time_dim,
    append,
)
from chirho.dynamical.internals.solver import simulate_trajectory
from chirho.dynamical.ops import Dynamics, State
from chirho.indexed.ops import IndexSet, gather, get_index_plates

T = TypeVar("T")


class LogTrajectory(Generic[T], pyro.poutine.messenger.Messenger):
    """
    An effect handler that logs the trajectory of a dynamical system at specified times. This is useful when
    interested in more than just the final state of the dynamical system. This can be used as below in conjunction
    with a specified solver backend, such as :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`.

    .. code-block:: python

            times = torch.linspace(0, 10, 100)
            with TorchDiffEq():
                with LogTrajectory(times) as trajectory_logger:
                    simulate(dynamics, initial_state, start_time, end_time)

    `trajectory_logger.trajectory` can be then be accessed to yield an object of type
    :ref:`State[T] <type-alias-State>`, but where
    each value in the mapping has an additional time dimension appended to the end. For example, if the shape of a state
    named `'x'` is `(3, 4)`, then the shape of `trajectory_logger.trajectory['x']` will be `(3, 4, 100)`.

    :param times: The times at which to log the trajectory.
    :type times: torch.Tensor

    :param is_traced: Whether to trace the trajectory. If True and executed within the context of a pyro trace,
        the trajectory will appear in the trace. Requires T to be a torch.Tensor.

    """

    trajectory: State[T]
    _trajectory: State[T]

    def __init__(self, times: torch.Tensor, is_traced: bool = False):
        self.times = times
        self._trajectory: State[T] = dict()
        self.is_traced = is_traced

        # Require that the times are sorted. This is required by the index masking we do below.
        if not torch.all(self.times[1:] > self.times[:-1]):
            raise ValueError("The passed times must be sorted.")

        super().__init__()

    def _pyro_post_simulate(self, msg: dict) -> None:
        initial_state: State[T] = msg["args"][1]
        start_time = msg["args"][2]

        if start_time == self.times[0]:
            # If we're starting at the beginning of the timespan, we need to log the initial state.
            # LogTrajectory's simulate_point will log only timepoints that are greater than the start_time of each
            # simulate_point call, which can occur multiple times in a single simulate call when there
            # are interruptions.
            self._trajectory: State[T] = append(
                _unsqueeze_time_dim(initial_state), self._trajectory
            )

        # Clear the internal trajectory so that we don't keep appending to it on subsequent simulate calls.
        self.trajectory: State[T] = self._trajectory
        self._trajectory: State[T] = type(initial_state)()

        if self.is_traced:
            for name, value in self.trajectory.items():
                assert isinstance(value, torch.Tensor)
                # This adds the trajectory to the trace so that it can be accessed later.
                pyro.deterministic(name, value)

    def _pyro_simulate_point(self, msg) -> None:
        # Turn a simulate that returns a state into a simulate that returns a trajectory at each of the logging_times
        dynamics: Dynamics[T] = msg["args"][0]
        initial_state: State[T] = msg["args"][1]
        start_time = msg["args"][2]
        end_time = msg["args"][3]

        if not self._trajectory:  # get type right
            self._trajectory = type(initial_state)()

        filtered_timespan = self.times[
            (self.times > start_time) & (self.times <= end_time)
        ]
        timespan = torch.concat(
            (start_time.unsqueeze(-1), filtered_timespan, end_time.unsqueeze(-1))
        )

        with pyro.poutine.messenger.block_messengers(lambda m: m is self):
            trajectory: State[T] = simulate_trajectory(
                dynamics, initial_state, timespan, **msg["kwargs"]
            )

        # TODO support dim != -1
        idx_name = "__time"
        name_to_dim = {}
        for k, f in get_index_plates().items():
            assert f.dim is not None
            name_to_dim[k] = f.dim - 1
        name_to_dim[idx_name] = -1

        if len(timespan) > 2:
            part_idx = IndexSet(**{idx_name: set(range(1, len(timespan) - 1))})
            new_part: State[T] = gather(trajectory, part_idx, name_to_dim=name_to_dim)
            self._trajectory: State[T] = append(self._trajectory, new_part)

        final_idx = IndexSet(**{idx_name: {len(timespan) - 1}})
        msg["value"] = _squeeze_time_dim(
            gather(trajectory, final_idx, name_to_dim=name_to_dim)
        )
        msg["done"] = True



from .components import (  # noqa: F401
    ExtractSupports,
    random_intervention,
    sufficiency_intervention,
    undo_split,
)
from .explanation import SearchForExplanation, SplitSubsets  # noqa: F401
from .preemptions import Preemptions  # noqa: F401



import functools
from typing import Any, Callable, Mapping, Protocol, TypeVar

import pyro
from typing_extensions import ParamSpec

from chirho.observational.ops import Observation

P = ParamSpec("P")
Q = ParamSpec("Q")
S = TypeVar("S", covariant=True)
T = TypeVar("T")

Point = Mapping[str, Observation[T]]


class Functional(Protocol[P, S]):
    def __call__(
        self, __model: Callable[P, Any], *models: Callable[P, Any]
    ) -> Callable[P, S]: ...


def influence_fn(
    functional: Functional[P, S],
    *points: Point[T],
    pointwise_influence: bool = True,
) -> Functional[P, S]:
    """
    Returns a new functional that computes the efficient influence function for ``functional``
    at the given ``points`` with respect to the parameters of its probabilistic program arguments.

    :param functional: model summary of interest, which is a function of ``model``
    :param points: points for each input to ``functional`` at which to compute the efficient influence function
    :return: functional that computes the efficient influence function for ``functional`` at ``points``

    **Example usage**:

        .. code-block:: python

            import pyro
            import pyro.distributions as dist
            import torch

            from chirho.observational.handlers.predictive import PredictiveModel
            from chirho.robust.handlers.estimators import MonteCarloInfluenceEstimator
            from chirho.robust.ops import influence_fn

            pyro.settings.set(module_local_params=True)


            class SimpleModel(pyro.nn.PyroModule):
                def forward(self):
                    a = pyro.sample("a", dist.Normal(0, 1))
                    with pyro.plate("data", 3, dim=-1):
                        b = pyro.sample("b", dist.Normal(a, 1))
                        return pyro.sample("y", dist.Normal(b, 1))


            class SimpleGuide(torch.nn.Module):
                def __init__(self):
                    super().__init__()
                    self.loc_a = torch.nn.Parameter(torch.rand(()))
                    self.loc_b = torch.nn.Parameter(torch.rand((3,)))

                def forward(self):
                    a = pyro.sample("a", dist.Normal(self.loc_a, 1))
                    with pyro.plate("data", 3, dim=-1):
                        b = pyro.sample("b", dist.Normal(self.loc_b, 1))
                        return {"a": a, "b": b}


            class SimpleFunctional(torch.nn.Module):
                def __init__(self, model, num_monte_carlo=1000):
                    super().__init__()
                    self.model = model
                    self.num_monte_carlo = num_monte_carlo

                def forward(self):
                    with pyro.plate("monte_carlo_functional", size=self.num_monte_carlo, dim=-2):
                        model_samples = pyro.poutine.trace(self.model).get_trace()
                    return model_samples.nodes["b"]["value"].mean(axis=0)


            model = SimpleModel()
            guide = SimpleGuide()
            predictive = pyro.infer.Predictive(
                model, guide=guide, num_samples=10, return_sites=["y"]
            )
            points = predictive()

            influence = influence_fn(
                SimpleFunctional,
                points,
            )(PredictiveModel(model, guide))

            with MonteCarloInfluenceEstimator(num_samples_inner=1000, num_samples_outer=1000):
                with torch.no_grad():  # Avoids memory leak (see notes below)
                    influence()

    .. note::

        * There are memory leaks when calling this function multiple times due to ``torch.func``.
          See issue:
          https://github.com/BasisResearch/chirho/issues/516.
          To avoid this issue, use ``torch.no_grad()`` as shown in the example above.

    """

    def influence_functional(
        *models: Callable[P, Any],
    ) -> Callable[P, S]:
        """
        Functional representing the efficient influence function of ``functional`` at ``points`` .

        :param models: Python callables containing Pyro primitives.
        :return: efficient influence function for ``functional`` evaluated at ``model`` and ``points``
        """

        @pyro.poutine.runtime.effectful(type="influence")
        def _influence(*model_args: P.args, **model_kwargs: P.kwargs) -> S:
            """
            Evaluates the efficient influence function for ``functional`` at each
            point in ``points``.

            :return: efficient influence function evaluated at each point in ``points`` or averaged
            """
            raise NotImplementedError(
                "Evaluating the `influence` induced by an `influence_fn` requires either "
                "(i) an approximation method such as `MonteCarloInfluenceEstimator`"
                "or (ii) a custom handler for the specific model and functional."
            )

        # This small amount of indirection allows any enclosing handlers to maintain a reference to
        # the `models`, `functional`, `points`, and `pointwise_influence` induced by the higher order `influence_fn`.
        return functools.partial(
            _influence,
            models=models,
            functional=functional,
            points=points,
            pointwise_influence=pointwise_influence,
        )

    return influence_functional



import functools
from typing import Optional, Tuple, TypeVar

import pyro

from chirho.indexed.ops import IndexSet, cond_n
from chirho.interventional.ops import Intervention, intervene

S = TypeVar("S")
T = TypeVar("T")


@pyro.poutine.runtime.effectful(type="preempt")
@functools.partial(pyro.poutine.block, hide_types=["intervene"])
def preempt(
    obs: T, acts: Tuple[Intervention[T], ...], case: Optional[S] = None, **kwargs
) -> T:
    """
    Effectful primitive operation for "preempting" values in a probabilistic program.

    Unlike the counterfactual operation :func:`~chirho.counterfactual.ops.split`,
    which returns multiple values concatenated along a new axis
    via the operation :func:`~chirho.indexed.ops.scatter`,
    :func:`preempt` returns a single value determined by the argument ``case``
    via :func:`~chirho.indexed.ops.cond`.

    In a probabilistic program, a :func:`preempt` call induces a mixture distribution
    over downstream values, whereas :func:`split` would induce a joint distribution.

    :param obs: The observed value.
    :param acts: The interventions to apply.
    :param case: The case to select.
    """
    if case is None:
        return obs

    name = kwargs.get("name", None)
    act_values = {IndexSet(**{name: {0}}): obs}
    for i, act in enumerate(acts):
        act_values[IndexSet(**{name: {i + 1}})] = intervene(obs, act, **kwargs)

    return cond_n(act_values, case, event_dim=kwargs.get("event_dim", 0))



import functools
from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple, TypeVar

import pyro
import torch
import torchdiffeq  # type: ignore

from chirho.dynamical.internals._utils import _squeeze_time_dim, _var_order
from chirho.dynamical.internals.solver import Interruption, simulate_point
from chirho.dynamical.ops import Dynamics, State
from chirho.indexed.ops import IndexSet, gather, get_index_plates

S = TypeVar("S")
T = TypeVar("T")


class TorchdiffeqRuntimeCheck(pyro.poutine.messenger.Messenger):
    def _pyro_sample(self, msg):
        raise ValueError(
            "TorchDiffEq only supports ODE models, and thus does not allow `pyro.sample` calls."
        )


def torchdiffeq_check_dynamics(
    dynamics: Dynamics[torch.Tensor],
    initial_state: State[torch.Tensor],
    start_time: torch.Tensor,
    end_time: torch.Tensor,
    **kwargs,
) -> None:
    with TorchdiffeqRuntimeCheck():
        dynamics(initial_state)


def _deriv(
    dynamics: Dynamics[torch.Tensor],
    var_order: Tuple[str, ...],
    time: torch.Tensor,
    state: Tuple[torch.Tensor, ...],
) -> Tuple[torch.Tensor, ...]:
    assert "t" not in var_order, "variable name t is reserved for time"
    env: State[torch.Tensor] = dict(zip(var_order + ("t",), state + (time,)))
    ddt: State[torch.Tensor] = dynamics(env)
    return tuple(ddt.get(var, torch.tensor(0.0)) for var in var_order)


def _torchdiffeq_ode_simulate_inner(
    dynamics: Dynamics[torch.Tensor],
    initial_state: State[torch.Tensor],
    timespan,
    **odeint_kwargs,
) -> State[torch.Tensor]:
    var_order = _var_order(frozenset(initial_state.keys()))  # arbitrary, but fixed

    diff = timespan[:-1] < timespan[1:]

    # We should only encounter collisions at the beginning or end of the timespan.
    if not torch.all(diff[1:-1]):
        raise ValueError(
            "elements of timespan must be strictly increasing, except at endpoints where interruptions can occur."
        )

    # Add a leading "true" to diff for masking, as we've excluded the first element.
    timespan_ = timespan[torch.cat((torch.tensor([True]), diff))]

    # time_dim is set to -1 by convention.
    # TODO: change this when time dim is allowed to vary.
    time_dim = -1

    if torch.any(diff):
        solns = _batched_odeint(  # torchdiffeq.odeint(
            functools.partial(_deriv, dynamics, var_order),
            tuple(initial_state[v] for v in var_order),
            timespan_,
            **odeint_kwargs,
        ).y
    else:
        solns = tuple(initial_state[v].unsqueeze(time_dim) for v in var_order)

    # As we've already asserted that collisions only happen at the beginning or end of the timespan, we can just
    #  concatenate the initial and final states to get the full trajectory if there are collisions.
    if not diff[0].item():
        solns = tuple(
            torch.cat((s[..., 0].unsqueeze(time_dim), s), dim=time_dim) for s in solns
        )
    if not diff[-1].item() and len(diff) > 1:
        solns = tuple(
            torch.cat((s, s[..., -1].unsqueeze(time_dim)), dim=time_dim) for s in solns
        )

    return type(initial_state)(**dict(zip(var_order, solns)))


@dataclass
class OdeintResult:
    y: Tuple[torch.Tensor, ...]
    event: Optional[torch.Tensor]


def _batched_odeint(
    func: Callable[[torch.Tensor, Tuple[torch.Tensor, ...]], Tuple[torch.Tensor, ...]],
    y0: Tuple[torch.Tensor, ...],
    t: torch.Tensor,
    *,
    event_fn=None,
    **odeint_kwargs,
) -> OdeintResult:
    """
    Vectorized torchdiffeq.odeint.
    """
    # TODO support event_dim > 0
    event_dim = 0  # assume states are batches of values of rank event_dim

    # Call deriv once to get the shape of the output.
    t0 = t.unsqueeze(0) if t.ndim == 0 else t[..., 0]
    dy0 = func(t0, y0)

    y0_batch_shape = torch.broadcast_shapes(
        *(dy0_.shape[: len(dy0_.shape) - event_dim] for dy0_ in dy0)
    )

    y0_expanded = tuple(
        # y0_[(None,) * (len(y0_batch_shape) - (len(y0_.shape) - event_dim)) + (...,)]
        y0_.expand(y0_batch_shape + y0_.shape[len(y0_.shape) - event_dim :])
        for y0_ in y0
    )

    if event_fn is not None:
        event_t, yt_raw = torchdiffeq.odeint_event(
            func, y0_expanded, t, event_fn=event_fn, **odeint_kwargs
        )
    else:
        yt_raw = torchdiffeq.odeint(func, y0_expanded, t, **odeint_kwargs)

    yt = tuple(
        torch.transpose(
            yt_[(..., None) + yt_.shape[len(yt_.shape) - event_dim :]],
            -len(yt_.shape) - 1,
            -1 - event_dim,
        )[0]
        for yt_ in yt_raw
    )
    return OdeintResult(yt, event_t if event_fn is not None else None)


def torchdiffeq_simulate_point(
    dynamics: Dynamics[torch.Tensor],
    initial_state: State[torch.Tensor],
    start_time: torch.Tensor,
    end_time: torch.Tensor,
    **kwargs,
) -> State[torch.Tensor]:
    timespan = torch.stack((start_time, end_time))
    trajectory = _torchdiffeq_ode_simulate_inner(
        dynamics, initial_state, timespan, **kwargs
    )

    # TODO support dim != -1
    idx_name = "__time"
    name_to_dim = {}
    for k, f in get_index_plates().items():
        assert f.dim is not None
        name_to_dim[k] = f.dim - 1
    name_to_dim[idx_name] = -1

    final_idx = IndexSet(**{idx_name: {len(timespan) - 1}})
    final_state_traj = gather(trajectory, final_idx, name_to_dim=name_to_dim)
    final_state = _squeeze_time_dim(final_state_traj)
    return final_state


def torchdiffeq_simulate_trajectory(
    dynamics: Dynamics[torch.Tensor],
    initial_state: State[torch.Tensor],
    timespan: torch.Tensor,
    **kwargs,
) -> State[torch.Tensor]:
    return _torchdiffeq_ode_simulate_inner(dynamics, initial_state, timespan, **kwargs)


def _torchdiffeq_get_next_interruptions(
    dynamics: Dynamics[torch.Tensor],
    start_state: State[torch.Tensor],
    start_time: torch.Tensor,
    interruptions: List[Interruption[torch.Tensor]],
    **kwargs,
) -> Tuple[Tuple[Interruption[torch.Tensor], ...], torch.Tensor]:
    # special case: static interruptions
    from chirho.dynamical.handlers.interruption import StaticEvent

    assert len(interruptions) > 0, "should have at least one interruption here"
    if all(isinstance(i.predicate, StaticEvent) for i in interruptions):
        next_static_interruption = min(
            interruptions, key=lambda i: getattr(i.predicate, "time")
        )
        assert isinstance(next_static_interruption.predicate, StaticEvent)
        return (next_static_interruption,), next_static_interruption.predicate.time

    var_order = _var_order(frozenset(start_state.keys()))  # arbitrary, but fixed

    # Create the event function combining all dynamic events and the terminal (next) static interruption.
    combined_event_f = torchdiffeq_combined_event_f(interruptions, var_order)

    # Simulate to the event execution.
    solns = _batched_odeint(  # torchdiffeq.odeint_event(
        functools.partial(_deriv, dynamics, var_order),
        tuple(start_state[v] for v in var_order),
        start_time,
        event_fn=combined_event_f,
        **kwargs,
    )
    event_time = solns.event
    assert event_time is not None
    event_solutions = solns.y

    # event_state has both the first and final state of the interrupted simulation. We just want the last.
    event_solution: Tuple[torch.Tensor, ...] = tuple(
        s[..., -1] for s in event_solutions
    )  # TODO support event_dim > 0

    # Check which event(s) fired, and put the triggered events in a list.
    # TODO support batched outputs of event functions
    fired_mask = torch.isclose(
        combined_event_f(event_time, event_solution),
        torch.tensor(0.0),
        rtol=1e-02,
        atol=1e-03,
    ).reshape(-1)

    if not torch.any(fired_mask):
        # TODO AZ figure out the tolerance of the odeint_event function and use that above.
        raise RuntimeError(
            "The solve terminated but no element of the event function output was within "
            "tolerance of zero."
        )

    if len(fired_mask) != len(interruptions):
        raise RuntimeError(
            "The event function returned an unexpected number of events."
        )

    triggered_events = [de for de, fm in zip(interruptions, fired_mask) if fm]

    return (
        tuple(triggered_events),
        event_time,
    )


def torchdiffeq_simulate_to_interruption(
    interruptions: List[Interruption[torch.Tensor]],
    dynamics: Dynamics[torch.Tensor],
    initial_state: State[torch.Tensor],
    start_time: torch.Tensor,
    end_time: torch.Tensor,
    **kwargs,
) -> Tuple[State[torch.Tensor], torch.Tensor, Optional[Interruption[torch.Tensor]]]:
    assert len(interruptions) > 0, "should have at least one interruption here"

    next_interruptions, interruption_time = _torchdiffeq_get_next_interruptions(
        dynamics, initial_state, start_time, interruptions, **kwargs
    )

    # Choose one arbitrarily if there are multiple.
    next_interruption = next_interruptions[0]

    value: State[torch.Tensor] = simulate_point(
        dynamics, initial_state, start_time, interruption_time, **kwargs
    )
    return value, interruption_time, next_interruption


def torchdiffeq_combined_event_f(
    interruptions: List[Interruption[torch.Tensor]],
    var_order: Tuple[str, ...],
) -> Callable[[torch.Tensor, Tuple[torch.Tensor, ...]], torch.Tensor]:
    """
    Construct a combined event function from a list of dynamic interruptions

    :param interruptions: The dynamic interruptions.
    :return: The combined event function, taking in state and time, and returning a vector of floats. When any element
     of this vector is zero, the corresponding event terminates the simulation.
    """
    from chirho.dynamical.handlers.interruption import ZeroEvent

    assert all(isinstance(i.predicate, ZeroEvent) for i in interruptions)

    def combined_event_f(t: torch.Tensor, flat_state: Tuple[torch.Tensor, ...]):
        state: State[torch.Tensor] = dict(zip(var_order, flat_state))
        return torch.stack(
            torch.broadcast_tensors(
                *[getattr(di.predicate, "event_fn")(t, state) for di in interruptions]
            ),
            dim=-1,
        )

    return combined_event_f



from __future__ import annotations

from typing import TypeVar

import torch

from chirho.dynamical.internals.solver import Solver

S = TypeVar("S")
T = TypeVar("T")


class TorchDiffEq(Solver[torch.Tensor]):
    """
    A dynamical systems solver backend for ordinary differential equations using
    `torchdiffeq <https://github.com/rtqichen/torchdiffeq>`_.
    When used in conjunction with :func:`simulate <chirho.dynamical.ops.simulate>`, as below, this backend will take
    responsibility for simulating the dynamical system defined by the arguments
    to :func:`simulate <chirho.dynamical.ops.simulate>`

    .. code-block:: python

        with TorchDiffEq():
            simulate(dynamics, initial_state, start_time, end_time)

    Additional details on the arguments below can be found in the
    `torchdiffeq documentation <https://github.com/rtqichen/torchdiffeq#keyword-arguments>`_

    :param rtol: The relative tolerance for the solver.
    :type rtol: float

    :param atol: The absolute tolerance for the solver.
    :type atol: float

    :param method: The solver method to use.
    :type method: str

    :param options: Additional options to pass to the solver.
    :type options: dict

    """

    def __init__(self, rtol=1e-7, atol=1e-9, method=None, options=None):
        self.rtol = rtol
        self.atol = atol
        self.method = method
        self.options = options
        self.odeint_kwargs = {
            "rtol": rtol,
            "atol": atol,
            "method": method,
            "options": options,
        }
        super().__init__()

    def _pyro_simulate_point(self, msg) -> None:
        from chirho.dynamical.internals.backends.torchdiffeq import (
            torchdiffeq_simulate_point,
        )

        dynamics, initial_state, start_time, end_time = msg["args"]
        msg["kwargs"].update(self.odeint_kwargs)

        msg["value"] = torchdiffeq_simulate_point(
            dynamics, initial_state, start_time, end_time, **msg["kwargs"]
        )
        msg["done"] = True

    def _pyro_simulate_trajectory(self, msg) -> None:
        from chirho.dynamical.internals.backends.torchdiffeq import (
            torchdiffeq_simulate_trajectory,
        )

        dynamics, initial_state, timespan = msg["args"]
        msg["kwargs"].update(self.odeint_kwargs)

        msg["value"] = torchdiffeq_simulate_trajectory(
            dynamics, initial_state, timespan, **msg["kwargs"]
        )
        msg["done"] = True

    def _pyro_simulate_to_interruption(self, msg) -> None:
        from chirho.dynamical.internals.backends.torchdiffeq import (
            torchdiffeq_simulate_to_interruption,
        )

        interruptions, dynamics, initial_state, start_time, end_time = msg["args"]
        msg["kwargs"].update(self.odeint_kwargs)
        msg["value"] = torchdiffeq_simulate_to_interruption(
            interruptions,
            dynamics,
            initial_state,
            start_time,
            end_time,
            **msg["kwargs"],
        )
        msg["done"] = True

    def _pyro_check_dynamics(self, msg) -> None:
        from chirho.dynamical.internals.backends.torchdiffeq import (
            torchdiffeq_check_dynamics,
        )

        dynamics, initial_state, start_time, end_time = msg["args"]
        msg["kwargs"].update(self.odeint_kwargs)

        torchdiffeq_check_dynamics(
            dynamics, initial_state, start_time, end_time, **msg["kwargs"]
        )
        msg["done"] = True



import collections.abc
import dataclasses
import functools
import typing
from typing import FrozenSet, Generic, Tuple, TypeVar

import pyro
import torch

from chirho.dynamical.ops import State

S = TypeVar("S")
T = TypeVar("T")


@functools.singledispatch
def append(fst, rest: T) -> T:
    raise NotImplementedError(f"append not implemented for type {type(fst)}.")


@append.register(collections.abc.Mapping)
def _append_trajectory(traj1: State[T], traj2: State[T]) -> State[T]:
    if len(traj1.keys()) == 0:
        return traj2

    if len(traj2.keys()) == 0:
        return traj1

    if traj1.keys() != traj2.keys():
        raise ValueError(
            f"Trajectories must have the same keys to be appended, but got {traj1.keys()} and {traj2.keys()}."
        )

    return type(traj1)(**{k: append(traj1[k], traj2[k]) for k in traj1.keys()})


@append.register(torch.Tensor)
def _append_tensor(prev_v: torch.Tensor, curr_v: torch.Tensor) -> torch.Tensor:
    time_dim = -1  # TODO generalize to nontrivial event_shape
    batch_shape = torch.broadcast_shapes(prev_v.shape[:-1], curr_v.shape[:-1])
    prev_v = prev_v.expand(*batch_shape, *prev_v.shape[-1:])
    curr_v = curr_v.expand(*batch_shape, *curr_v.shape[-1:])
    return torch.cat([prev_v, curr_v], dim=time_dim)


@functools.lru_cache
def _var_order(varnames: FrozenSet[str]) -> Tuple[str, ...]:
    return tuple(sorted(varnames))


@functools.singledispatch
def _squeeze_time_dim(state_or_traj):
    raise NotImplementedError(
        f"_squeeze_time_dim not implemented for type {type(state_or_traj)}."
    )


@_squeeze_time_dim.register(collections.abc.Mapping)
def _squeeze_time_dim_trajectory(traj: State[T]) -> State[T]:
    return type(traj)(**{k: _squeeze_time_dim(traj[k]) for k in traj.keys()})


@_squeeze_time_dim.register(torch.Tensor)
def _squeeze_time_dim_tensor(state: torch.Tensor) -> torch.Tensor:
    return state.squeeze(-1)


@functools.singledispatch
def _unsqueeze_time_dim(state_or_traj):
    raise NotImplementedError(
        f"_unsqueeze_time_dim not implemented for type {type(state_or_traj)}."
    )


@_unsqueeze_time_dim.register(collections.abc.Mapping)
def _unsqueeze_time_dim_state(state: State[T]) -> State[T]:
    return type(state)(**{k: _unsqueeze_time_dim(state[k]) for k in state.keys()})


@_unsqueeze_time_dim.register(torch.Tensor)
def _unsqueeze_time_dim_tensor(state: torch.Tensor) -> torch.Tensor:
    return state.unsqueeze(-1)


class ShallowMessenger(pyro.poutine.messenger.Messenger):
    """
    Base class for so-called "shallow" effect handlers that uninstall themselves
    after handling a single operation.

    .. warning::

        Does not support post-processing or overriding generic ``_process_message``
    """

    used: bool

    def __enter__(self):
        self.used = False
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        if self in pyro.poutine.runtime._PYRO_STACK:
            super().__exit__(exc_type, exc_value, traceback)

    @typing.final
    def _process_message(self, msg: pyro.poutine.messenger.Message) -> None:
        if not self.used and hasattr(self, f"_pyro_{msg['type']}"):
            self.used = True
            super()._process_message(msg)

            prev_cont = msg["continuation"]

            def cont(msg: pyro.poutine.messenger.Message) -> None:
                ix = pyro.poutine.runtime._PYRO_STACK.index(self)
                pyro.poutine.runtime._PYRO_STACK.pop(ix)
                if prev_cont is not None:
                    prev_cont(msg)

            msg["continuation"] = cont

    @typing.final
    def _postprocess_message(self, msg: pyro.poutine.messenger.Message) -> None:
        if hasattr(self, f"_pyro_post_{msg['type']}"):
            raise NotImplementedError("ShallowHandler does not support postprocessing")


@dataclasses.dataclass(order=True)
class Prioritized(Generic[T]):
    priority: float
    item: T = dataclasses.field(compare=False)



import collections
import math
import typing
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar

import pyro
import torch
from typing_extensions import ParamSpec

from chirho.indexed.handlers import IndexPlatesMessenger
from chirho.indexed.ops import get_index_plates
from chirho.observational.handlers.predictive import BatchedLatents, BatchedObservations
from chirho.robust.ops import Point

pyro.settings.set(module_local_params=True)

P = ParamSpec("P")
Q = ParamSpec("Q")
S = TypeVar("S")
T = TypeVar("T")


def get_importance_traces(
    model: Callable[P, Any],
    guide: Optional[Callable[P, Any]] = None,
) -> Callable[P, Tuple[pyro.poutine.Trace, pyro.poutine.Trace]]:
    """
    Thin functional wrapper around :func:`~pyro.infer.enum.get_importance_trace`
    that cleans up the original interface to avoid unnecessary arguments
    and efficiently supports using the prior in a model as a default guide.

    :param model: Model to run.
    :param guide: Guide to run. If ``None``, use the prior in ``model`` as a guide.
    :returns: A function that takes the same arguments as ``model`` and ``guide`` and returns
        a tuple of importance traces ``(model_trace, guide_trace)``.
    """

    def _fn(
        *args: P.args, **kwargs: P.kwargs
    ) -> Tuple[pyro.poutine.Trace, pyro.poutine.Trace]:
        if guide is not None:
            model_trace, guide_trace = pyro.infer.enum.get_importance_trace(
                "flat", math.inf, model, guide, args, kwargs
            )
            return model_trace, guide_trace
        else:  # use prior as default guide, but don't run model twice
            model_trace, _ = pyro.infer.enum.get_importance_trace(
                "flat", math.inf, model, lambda *_, **__: None, args, kwargs
            )

            guide_trace = model_trace.copy()
            for name, node in list(guide_trace.nodes.items()):
                if node["type"] != "sample":
                    del model_trace.nodes[name]
                elif pyro.poutine.util.site_is_factor(node) or node["is_observed"]:
                    del guide_trace.nodes[name]
            return model_trace, guide_trace

    return _fn


class BatchedNMCLogMarginalLikelihood(Generic[P, T], torch.nn.Module):
    r"""
    Approximates the log marginal likelihood induced by ``model`` and ``guide``
    using importance sampling at an arbitrary batch of :math:`N`
    points :math:`\{x_n\}_{n=1}^N`.

    .. math::
        \log \left(\frac{1}{M} \sum_{m=1}^M \frac{p(x_n \mid \theta_m) p(\theta_m) )}{q_{\phi}(\theta_m)} \right),
        \quad \theta_m \sim q_{\phi}(\theta),

    where :math:`q_{\phi}(\theta)` is the guide, and :math:`p(x_n \mid \theta_m) p(\theta_m)`
    is the model joint density of the data and the latents sampled from the guide.

    :param model: Python callable containing Pyro primitives.
    :type model: torch.nn.Module
    :param guide: Python callable containing Pyro primitives.
        Must only contain continuous latent variables.
    :type guide: torch.nn.Module
    :param num_samples: Number of Monte Carlo draws :math:`M`
        used to approximate marginal distribution, defaults to 1
    :type num_samples: int, optional
    """

    model: Callable[P, Any]
    guide: Optional[Callable[P, Any]]
    num_samples: int

    def __init__(
        self,
        model: torch.nn.Module,
        guide: Optional[torch.nn.Module] = None,
        *,
        num_samples: int = 1,
        max_plate_nesting: Optional[int] = None,
        data_plate_name: str = "__particles_data",
        mc_plate_name: str = "__particles_mc",
    ):
        super().__init__()
        self.model = model
        self.guide = guide
        self.num_samples = num_samples
        self._first_available_dim = (
            -max_plate_nesting - 1 if max_plate_nesting is not None else None
        )
        self._data_plate_name = data_plate_name
        self._mc_plate_name = mc_plate_name

    def forward(
        self, data: Point[T], *args: P.args, **kwargs: P.kwargs
    ) -> torch.Tensor:
        """
        Computes the log predictive likelihood of ``data`` given ``model`` and ``guide``.

        :param data: Dictionary of observations.
        :type data: Point[T]
        :return: Log marginal likelihood at each datapoint.
        :rtype: torch.Tensor
        """
        get_nmc_traces = get_importance_traces(self.model, self.guide)

        with IndexPlatesMessenger(first_available_dim=self._first_available_dim):
            with BatchedLatents(self.num_samples, name=self._mc_plate_name):
                with BatchedObservations(data, name=self._data_plate_name):
                    model_trace, guide_trace = get_nmc_traces(*args, **kwargs)
            index_plates = get_index_plates()

        plate_name_to_dim = collections.OrderedDict(
            (p, index_plates[p])
            for p in [self._mc_plate_name, self._data_plate_name]
            if p in index_plates
        )
        plate_frames = set(plate_name_to_dim.values())

        log_weights = typing.cast(torch.Tensor, 0.0)
        for site in model_trace.nodes.values():
            if site["type"] != "sample":
                continue
            site_log_prob = site["log_prob"]
            for f in site["cond_indep_stack"]:
                if f.dim is not None and f not in plate_frames:
                    site_log_prob = site_log_prob.sum(f.dim, keepdim=True)
            log_weights = log_weights + site_log_prob

        for site in guide_trace.nodes.values():
            if site["type"] != "sample":
                continue
            site_log_prob = site["log_prob"]
            for f in site["cond_indep_stack"]:
                if f.dim is not None and f not in plate_frames:
                    site_log_prob = site_log_prob.sum(f.dim, keepdim=True)
            log_weights = log_weights - site_log_prob

        # sum out particle dimension and discard
        if self._mc_plate_name in index_plates:
            dim = plate_name_to_dim[self._mc_plate_name].dim
            assert dim is not None
            log_weights = torch.logsumexp(
                log_weights, dim=dim, keepdim=True
            ) - math.log(self.num_samples)
            plate_name_to_dim.pop(self._mc_plate_name)

        # move data plate dimension to the left
        for name in reversed(plate_name_to_dim.keys()):
            dim = plate_name_to_dim[name].dim
            assert dim is not None
            log_weights = torch.transpose(
                log_weights[None], -len(log_weights.shape) - 1, dim
            )

        # pack log_weights by squeezing out rightmost dimensions
        for _ in range(len(log_weights.shape) - len(plate_name_to_dim)):
            log_weights = log_weights.squeeze(-1)

        return log_weights



from __future__ import annotations

import collections
import functools
from typing import Callable, Dict, Generic, Hashable, Mapping, Optional, TypeVar, Union

import pyro
import torch

from chirho.interventional.ops import (
    AtomicIntervention,
    CompoundIntervention,
    intervene,
)

K = TypeVar("K")
T = TypeVar("T")


@intervene.register(int)
@intervene.register(float)
@intervene.register(bool)
@intervene.register(torch.Tensor)
@pyro.poutine.runtime.effectful(type="intervene")
def _intervene_atom(
    obs, act: Optional[AtomicIntervention[T]] = None, *, event_dim: int = 0, **kwargs
) -> T:
    """
    Intervene on an atomic value in a probabilistic program.
    """
    if act is None:
        return obs
    elif callable(act):
        act = act(obs)
        return act[-1] if isinstance(act, tuple) else act
    elif isinstance(act, tuple):
        return act[-1]
    return act


@intervene.register(pyro.distributions.Distribution)
@pyro.poutine.runtime.effectful(type="intervene")
def _intervene_atom_distribution(
    obs: pyro.distributions.Distribution,
    act: Optional[AtomicIntervention[pyro.distributions.Distribution]] = None,
    **kwargs,
) -> pyro.distributions.Distribution:
    """
    Intervene on a distribution in a probabilistic program.
    """
    if act is None:
        return obs
    elif callable(act) and not isinstance(act, pyro.distributions.Distribution):
        act = act(obs)
        return act[-1] if isinstance(act, tuple) else act
    elif isinstance(act, tuple):
        return act[-1]
    return act


@intervene.register(dict)
def _dict_intervene(
    obs: Dict[K, T],
    act: Union[Dict[K, AtomicIntervention[T]], Callable[[Dict[K, T]], Dict[K, T]]],
    **kwargs,
) -> Dict[K, T]:

    if callable(act):
        return _dict_intervene_callable(obs, act, **kwargs)

    result: Dict[K, T] = {}
    for k in obs.keys():
        result[k] = intervene(obs[k], act[k] if k in act else None, **kwargs)
    return result


@pyro.poutine.runtime.effectful(type="intervene")
def _dict_intervene_callable(
    obs: Dict[K, T], act: Callable[[Dict[K, T]], Dict[K, T]], **kwargs
) -> Dict[K, T]:
    return act(obs)


@intervene.register
def _intervene_callable(
    obs: collections.abc.Callable,
    act: Optional[CompoundIntervention[T]] = None,
    **call_kwargs,
) -> Callable[..., T]:
    if act is None:
        return obs
    elif callable(act):

        @functools.wraps(obs)
        def _intervene_callable_wrapper(*args, **kwargs):
            return intervene(obs(*args, **kwargs), act(*args, **kwargs), **call_kwargs)

        return _intervene_callable_wrapper
    return Interventions(actions=act)(obs)


class Interventions(Generic[T], pyro.poutine.messenger.Messenger):
    """
    Intervene on values in a probabilistic program.

    :class:`DoMessenger` is an effect handler that intervenes at specified sample sites
    in a probabilistic program. This allows users to define programs without any
    interventional or causal semantics, and then to add those features later in the
    context of, for example, :class:`DoMessenger`. This handler uses :func:`intervene`
    internally and supports the same types of interventions.
    """

    def __init__(self, actions: Mapping[Hashable, AtomicIntervention[T]]):
        """
        :param actions: A mapping from names of sample sites to interventions.
        """
        self.actions = actions
        super().__init__()

    def _pyro_post_sample(self, msg):
        if msg["name"] not in self.actions or msg["infer"].get(
            "_do_not_intervene", None
        ):
            return

        msg["value"] = intervene(
            msg["value"],
            self.actions[msg["name"]],
            event_dim=len(msg["fn"].event_shape),
            name=msg["name"],
        )


if isinstance(pyro.poutine.handlers._make_handler(Interventions), tuple):
    do = pyro.poutine.handlers._make_handler(Interventions)[1]
else:

    @pyro.poutine.handlers._make_handler(Interventions)
    def do(fn: Callable, actions: Mapping[Hashable, AtomicIntervention[T]]): ...



from __future__ import annotations

import typing
from typing import Callable, Generic, Tuple, TypeVar, Union

import pyro
import pyro.contrib.autoname
import torch

from chirho.dynamical.handlers.trajectory import LogTrajectory
from chirho.dynamical.ops import Dynamics, State, on
from chirho.indexed.ops import cond
from chirho.interventional.ops import Intervention, intervene
from chirho.observational.ops import Observation, observe

R = Union[float, torch.Tensor]
S = TypeVar("S")
T = TypeVar("T")


class ZeroEvent(Generic[T]):
    """
    Class for creating event functions for use with :func:`~chirho.dynamical.ops.on`
    that trigger when a given scalar-valued function approaches and crosses ``0.``

    For example, to define an event handler that calls :func:`~chirho.interventional.ops.intervene`
    when the state variable `x` exceeds 10.0, we could use the following::

        @on(ZeroEvent(lambda time, state: state["x"] - 10.0)
        def callback(dynamics: Dynamics[T], state: State[T]) -> Tuple[Dynamics[T], State[T]]:
            return dynamics, intervene(state, {"x": 0.0})

    .. note:: some backends, such as :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`,
        only support event handler predicates specified via :class:`~ZeroEvent` ,
        not via arbitrary boolean-valued functions of the state.

    :param event_fn: A function that approaches and crosses 0.0 at the moment the event should be triggered.
    """

    event_fn: Callable[[R, State[T]], R]

    def __init__(self, event_fn: Callable[[R, State[T]], R]):
        self.event_fn = event_fn
        super().__init__()

    def __call__(self, state: State[T]) -> bool:
        return bool(self.event_fn(typing.cast(torch.Tensor, state["t"]), state) == 0.0)


class StaticEvent(Generic[T], ZeroEvent[T]):
    """
    Class for creating event functions for use with :func:`~chirho.dynamical.ops.on`
    that trigger at a specified time.

    For example, to define an event handler that calls :func:`~chirho.interventional.ops.intervene`
    at time 10.0, we could use the following::

        @on(StaticEvent(10.0))
        def callback(dynamics: Dynamics[T], state: State[T]) -> Tuple[Dynamics[T], State[T]]:
            return dynamics, intervene(state, {"x": 0.0})

    :param time: The time at which the event should be triggered.
    """

    time: torch.Tensor
    event_fn: Callable[[R, State[T]], R]

    def __init__(self, time: R):
        self.time = torch.as_tensor(time)
        super().__init__(
            lambda time, _: cond(0.0, self.time - time, case=time < self.time)
        )


def StaticInterruption(time: R):
    """
    A handler that will interrupt a simulation at a specified time, and then resume it afterward.
    Other handlers, such as :class:`~chirho.dynamical.handlers.interruption.StaticObservation`
    and :class:`~chirho.dynamical.handlers.interruption.StaticIntervention` subclass this handler to provide additional
    functionality.

    Won't generally be used by itself, but rather as a base class for other handlers.

    :param time: The time at which the simulation will be interrupted.
    """

    @on(StaticEvent(time))
    def callback(
        dynamics: Dynamics[T], state: State[T]
    ) -> Tuple[Dynamics[T], State[T]]:
        return dynamics, state

    return callback


def StaticObservation(time: R, observation: Observation[State[T]]):
    """
    This effect handler interrupts a simulation at a given time
    (as outlined by :class:`~chirho.dynamical.handlers.interruption.StaticInterruption`), and then applies
    a user-specified observation noise model to the state at that time. Typically, this noise model
    will be conditioned on some noisy observation of the state at that time. For a system involving a
    scalar state named `x`, it can be used like so:

    .. code-block:: python

        def observation(state: State[torch.Tensor]):
            pyro.sample("x_obs", dist.Normal(state["x"], 1.0))

        data = {"x_obs": torch.tensor(10.0)}
        obs = condition(data=data)(observation)
        with TorchDiffEq():
            with StaticObservation(time=2.9, observation=obs):
                result = simulate(dynamics, init_state, start_time, end_time)

    For details on other entities used above, see :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`,
    :func:`~chirho.dynamical.ops.simulate`, and :class:`~chirho.observational.handlers.condition`.

    :param time: The time at which the observation is made.
    :param observation: The observation noise model to apply to the state at the given time. Can be conditioned on data.
    """

    @on(StaticEvent(time))
    def callback(
        dynamics: Dynamics[T], state: State[T]
    ) -> Tuple[Dynamics[T], State[T]]:
        with pyro.contrib.autoname.scope(prefix=f"t={torch.as_tensor(time).item()}"):
            return dynamics, observe(state, observation)

    return callback


def StaticIntervention(time: R, intervention: Intervention[State[T]]):
    """
    This effect handler interrupts a simulation at a specified time, and applies an intervention to the state at that
    time. It can be used as below:

    .. code-block:: python

        intervention = {"x": torch.tensor(1.0)}
        with TorchDiffEq():
            with StaticIntervention(time=1.5, intervention=intervention):
                simulate(dynamics, init_state, start_time, end_time)

    For details on other entities used above, see :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`,
    :func:`~chirho.dynamical.ops.simulate`.

    :param time: The time at which the intervention is applied.
    :param intervention: The instantaneous intervention applied to the state when the event is triggered. The supplied
        intervention will be passed to :func:`~chirho.interventional.ops.intervene`, and as such can be any types
        supported by that function. This includes state dependent interventions specified by a function, such as
        `lambda state: {"x": state["x"] + 1.0}`.
    """

    @on(StaticEvent(time))
    def callback(
        dynamics: Dynamics[T], state: State[T]
    ) -> Tuple[Dynamics[T], State[T]]:
        return dynamics, intervene(state, intervention)

    return callback


def DynamicInterruption(event_fn: Callable[[R, State[T]], R]):
    """
    :param event_f: An event trigger function that approaches and returns 0.0 when the event should be triggered.
        This can be designed to trigger when the current state is "close enough" to some trigger state, or when an
        element of the state exceeds some threshold, etc. It takes both the current time and current state.
    """

    @on(ZeroEvent(event_fn))
    def callback(
        dynamics: Dynamics[T], state: State[T]
    ) -> Tuple[Dynamics[T], State[T]]:
        return dynamics, state

    return callback


def DynamicIntervention(
    event_fn: Callable[[R, State[T]], R], intervention: Intervention[State[T]]
):
    """
    This effect handler interrupts a simulation when the given dynamic event function returns 0.0, and
    applies an intervention to the state at that time. This works similarly to
    :class:`~chirho.dynamical.handlers.interruption.StaticIntervention`, but supports state-dependent trigger
    conditions for the intervention, as opposed to a static, time-dependent trigger condition.

    :param event_fn: An event trigger function that approaches and crosses 0.0 at the moment the intervention should be
     applied triggered. Upon triggering, the simulation is interrupted and the intervention is applied to the state. The
     event function takes both the current time and current state as arguments.
    :param intervention: The instantaneous intervention applied to the state when the event is triggered. The supplied
        intervention will be passed to :func:`~chirho.interventional.ops.intervene`, and as such can be any types
        supported by that function. This includes state dependent interventions specified by a function, such as
        `lambda state: {"x": state["x"] + 1.0}`.
    """

    @on(ZeroEvent(event_fn))
    def callback(
        dynamics: Dynamics[T], state: State[T]
    ) -> Tuple[Dynamics[T], State[T]]:
        return dynamics, intervene(state, intervention)

    return callback


class StaticBatchObservation(Generic[T], LogTrajectory[T]):
    """
    This effect handler behaves similarly to :class:`~chirho.dynamical.handlers.interruption.StaticObservation`,
    but does not interrupt the simulation. Instead, it uses :class:`~chirho.dynamical.handlers.trajectory.LogTrajectory`
    to log the trajectory of the system at specified times, and then applies an observation noise model to the
    logged trajectory. This is especially useful when one has many noisy observations of the system at different
    times, and/or does not want to incur the overhead of interrupting the simulation at each observation time.

    For a system involving a scalar state named `x`, it can be used like so:

    .. code-block:: python

        def observation(state: State[torch.Tensor]):
            pyro.sample("x_obs", dist.Normal(state["x"], 1.0))

        data = {"x_obs": torch.tensor([10., 20., 10.])}
        obs = condition(data=data)(observation)
        with TorchDiffEq():
            with StaticBatchObservation(times=torch.tensor([1.0, 2.0, 3.0]), observation=obs):
                result = simulate(dynamics, init_state, start_time, end_time)

    For details on other entities used above, see :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`,
    :func:`~chirho.dynamical.ops.simulate`, and :class:`~chirho.observational.handlers.condition`.

    :param times: The times at which the observations are made.
    :param observation: The observation noise model to apply to the logged trajectory. Can be conditioned on data.

    """

    observation: Observation[State[T]]

    def __init__(
        self,
        times: torch.Tensor,
        observation: Observation[State[T]],
        **kwargs,
    ):
        self.observation = observation
        super().__init__(times, **kwargs)

    def _pyro_post_simulate(self, msg: dict) -> None:
        super()._pyro_post_simulate(msg)
        self.trajectory = observe(self.trajectory, self.observation)



import functools
from typing import Any, Callable, Optional, TypeVar

import pyro
import torch
from typing_extensions import Concatenate, ParamSpec

from chirho.robust.internals.nmc import BatchedNMCLogMarginalLikelihood
from chirho.robust.internals.utils import (
    ParamDict,
    make_flatten_unflatten,
    make_functional_call,
    reset_rng_state,
)
from chirho.robust.ops import Point

P = ParamSpec("P")
Q = ParamSpec("Q")
S = TypeVar("S")
T = TypeVar("T")


def _flat_conjugate_gradient_solve(
    f_Ax: Callable[[torch.Tensor], torch.Tensor],
    b: torch.Tensor,
    *,
    cg_iters: Optional[int] = None,
    residual_tol: float = 1e-3,
) -> torch.Tensor:
    """
    Use Conjugate Gradient iteration to solve Ax = b. Demmel p 312.

    :param f_Ax: a function to compute matrix vector products over a batch
        of vectors ``x``.
    :type f_Ax: Callable[[torch.Tensor], torch.Tensor]
    :param b: batch of right hand sides of the equation to solve.
    :type b: torch.Tensor
    :param cg_iters: number of conjugate iterations to run, defaults to None
    :type cg_iters: Optional[int], optional
    :param residual_tol: tolerance for convergence, defaults to 1e-3
    :type residual_tol: float, optional
    :return: batch of solutions ``x*`` for equation Ax = b.
    :rtype: torch.Tensor

    .. note::

        Code is adapted from
        https://github.com/rlworkgroup/garage/blob/master/src/garage/torch/optimizers/conjugate_gradient_optimizer.py # noqa: E501

    """
    assert len(b.shape), "b must be a 2D matrix"

    if cg_iters is None:
        cg_iters = b.shape[1]
    else:
        cg_iters = min(cg_iters, b.shape[1])

    def _batched_dot(x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
        return (x1 * x2).sum(axis=-1)  # type: ignore

    def _batched_product(a: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
        return a.unsqueeze(0).t() * B

    p = b.clone()
    r = b.clone()
    x = torch.zeros_like(b)
    z = f_Ax(p)
    rdotr = _batched_dot(r, r)
    v = rdotr / _batched_dot(p, z)
    newrdotr = rdotr
    mu = newrdotr / rdotr
    zeros_xr = torch.zeros_like(x)
    for _ in range(cg_iters):
        not_converged = rdotr > residual_tol
        not_converged_broadcasted = not_converged.unsqueeze(0).t()
        z = torch.where(not_converged_broadcasted, f_Ax(p), z)
        v = torch.where(not_converged, rdotr / _batched_dot(p, z), v)
        x += torch.where(not_converged_broadcasted, _batched_product(v, p), zeros_xr)
        r -= torch.where(not_converged_broadcasted, _batched_product(v, z), zeros_xr)
        newrdotr = torch.where(not_converged, _batched_dot(r, r), newrdotr)
        mu = torch.where(not_converged, newrdotr / rdotr, mu)
        p = torch.where(not_converged_broadcasted, r + _batched_product(mu, p), p)
        rdotr = torch.where(not_converged, newrdotr, rdotr)
        if torch.all(~not_converged):
            return x
    return x


def conjugate_gradient_solve(f_Ax: Callable[[T], T], b: T, **kwargs) -> T:
    """
    Use Conjugate Gradient iteration to solve Ax = b.

    :param f_Ax: a function to compute matrix vector products over a batch
        of vectors ``x``.
    :type f_Ax: Callable[[T], T]
    :param b: batch of right hand sides of the equation to solve.
    :type b: T
    :return:  batch of solutions ``x*`` for equation Ax = b.
    :rtype: T
    """
    flatten, unflatten = make_flatten_unflatten(b)

    def f_Ax_flat(v: torch.Tensor) -> torch.Tensor:
        v_unflattened: T = unflatten(v)
        result_unflattened = f_Ax(v_unflattened)
        return flatten(result_unflattened)

    return unflatten(_flat_conjugate_gradient_solve(f_Ax_flat, flatten(b), **kwargs))


def make_empirical_fisher_vp(
    batched_func_log_prob: Callable[Concatenate[ParamDict, Point[T], P], torch.Tensor],
    log_prob_params: ParamDict,
    data: Point[T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Callable[[ParamDict], ParamDict]:
    r"""
    Returns a function that computes the empirical Fisher vector product for an arbitrary
    vector :math:`v` using only Hessian vector products via a batched version of
    Perlmutter's trick [1].

    .. math::

        -\frac{1}{N} \sum_{n=1}^N \nabla_{\phi}^2 \log \tilde{p}_{\phi}(x_n) v,

    where :math:`\phi` corresponds to ``log_prob_params``, :math:`\tilde{p}_{\phi}` denotes the
    predictive distribution ``log_prob``, and :math:`x_n` are the data points in ``data``.

    :param func_log_prob: computes the log probability of ``data`` given ``log_prob_params``
    :type func_log_prob: Callable[Concatenate[ParamDict, Point[T], P], torch.Tensor]
    :param log_prob_params: parameters of the predictive distribution
    :type log_prob_params: ParamDict
    :param data: data points
    :type data: Point[T]
    :param is_batched: if ``False``, ``func_log_prob`` is batched over ``data``
        using ``torch.func.vmap``. Otherwise, assumes ``func_log_prob`` is already batched
        over multiple data points. ``Defaults to False``.
    :type is_batched: bool, optional
    :return: a function that computes the empirical Fisher vector product for an arbitrary
        vector :math:`v`
    :rtype: Callable[[ParamDict], ParamDict]

    **Example usage**:

        .. code-block:: python

            import pyro
            import pyro.distributions as dist
            import torch

            from chirho.robust.internals.linearize import make_empirical_fisher_vp

            pyro.settings.set(module_local_params=True)


            class GaussianModel(pyro.nn.PyroModule):
                def __init__(self, cov_mat: torch.Tensor):
                    super().__init__()
                    self.register_buffer("cov_mat", cov_mat)

                def forward(self, loc):
                    pyro.sample(
                        "x", dist.MultivariateNormal(loc=loc, covariance_matrix=self.cov_mat)
                    )


            def gaussian_log_prob(params, data_point, cov_mat):
                with pyro.validation_enabled(False):
                    return dist.MultivariateNormal(
                        loc=params["loc"], covariance_matrix=cov_mat
                    ).log_prob(data_point["x"])


            v = torch.tensor([1.0, 0.0], requires_grad=False)
            loc = torch.ones(2, requires_grad=True)
            cov_mat = torch.ones(2, 2) + torch.eye(2)

            func_log_prob = gaussian_log_prob
            log_prob_params = {"loc": loc}
            N_monte_carlo = 10000
            data = pyro.infer.Predictive(GaussianModel(cov_mat), num_samples=N_monte_carlo)(loc)
            empirical_fisher_vp_func = make_empirical_fisher_vp(
                func_log_prob, log_prob_params, data, cov_mat=cov_mat
            )

            empirical_fisher_vp = empirical_fisher_vp_func({"loc": v})["loc"]

            # Closed form solution for the Fisher vector product
            # See "Multivariate normal distribution" in https://en.wikipedia.org/wiki/Fisher_information
            prec_matrix = torch.linalg.inv(cov_mat)
            true_vp = prec_matrix.mv(v)

            assert torch.all(torch.isclose(empirical_fisher_vp, true_vp, atol=0.1))


    **References**

    [1] `Fast Exact Multiplication by the Hessian`,
    Barak A. Pearlmutter, 1999.
    """
    N = data[next(iter(data))].shape[0]  # type: ignore
    mean_vector = 1 / N * torch.ones(N)

    def bound_batched_func_log_prob(params: ParamDict) -> torch.Tensor:
        return batched_func_log_prob(params, data, *args, **kwargs)

    def _empirical_fisher_vp(v: ParamDict) -> ParamDict:
        def jvp_fn(log_prob_params: ParamDict) -> torch.Tensor:
            return torch.func.jvp(
                bound_batched_func_log_prob, (log_prob_params,), (v,)
            )[1]

        # Perlmutter's trick
        vjp_fn = torch.func.vjp(jvp_fn, log_prob_params)[1]
        return vjp_fn(-1 * mean_vector)[0]  # Fisher = -E[Hessian]

    return _empirical_fisher_vp


def linearize(
    *models: Callable[P, Any],
    num_samples_outer: int,
    num_samples_inner: Optional[int] = None,
    max_plate_nesting: Optional[int] = None,
    cg_iters: Optional[int] = None,
    residual_tol: float = 1e-4,
    pointwise_influence: bool = True,
) -> Callable[Concatenate[Point[T], P], ParamDict]:
    r"""
    Returns the influence function associated with the parameters
    of a normalized probabilistic program ``model``. This function
    computes the following quantity at an arbitrary point :math:`x^{\prime}`:

    .. math::

        \left[-\frac{1}{N} \sum_{n=1}^N \nabla_{\phi}^2 \log \tilde{p}_{\phi}(x_n) \right]
        \nabla_{\phi} \log \tilde{p}_{\phi}(x^{\prime}), \quad
        \tilde{p}_{\phi}(x) = \int p_{\phi}(x, \theta) d\theta,

    where :math:`\phi` corresponds to ``log_prob_params``,
    :math:`p(x, \theta)` denotes the ``model``,
    :math:`\tilde{p}_{\phi}` denotes the predictive distribution ``log_prob`` induced
    from the ``model``, and :math:`\{x_n\}_{n=1}^N` are the
    data points drawn iid from the predictive distribution.

    :param model: Python callable containing Pyro primitives.
    :type model: Callable[P, Any]
    :param num_samples_outer: number of Monte Carlo samples to
        approximate Fisher information in :func:`make_empirical_fisher_vp`
    :type num_samples_outer: int
    :param num_samples_inner: number of Monte Carlo samples used in
        :class:`BatchedNMCLogPredictiveLikelihood`. Defaults to ``num_samples_outer**2``.
    :type num_samples_inner: Optional[int], optional
    :param max_plate_nesting: bound on max number of nested :func:`pyro.plate`
        contexts. Defaults to ``None``.
    :type max_plate_nesting: Optional[int], optional
    :param cg_iters: number of conjugate gradient steps used to
        invert Fisher information matrix, defaults to None
    :type cg_iters: Optional[int], optional
    :param residual_tol: tolerance used to terminate conjugate gradients
        early, defaults to 1e-4
    :type residual_tol: float, optional
    :param pointwise_influence: if ``True``, computes the influence function at each
        point in ``points``. If ``False``, computes the efficient influence averaged
        over ``points``. Defaults to True.
    :type pointwise_influence: bool, optional
    :return: the influence function associated with the parameters
    :rtype: Callable[Concatenate[Point[T], P], ParamDict]

    **Example usage**:

        .. code-block:: python

            import pyro
            import pyro.distributions as dist
            import torch

            from chirho.observational.handlers.predictive import PredictiveModel
            from chirho.robust.internals.linearize import linearize

            pyro.settings.set(module_local_params=True)


            class SimpleModel(pyro.nn.PyroModule):
                def forward(self):
                    a = pyro.sample("a", dist.Normal(0, 1))
                    with pyro.plate("data", 3, dim=-1):
                        b = pyro.sample("b", dist.Normal(a, 1))
                        return pyro.sample("y", dist.Normal(b, 1))


            class SimpleGuide(torch.nn.Module):
                def __init__(self):
                    super().__init__()
                    self.loc_a = torch.nn.Parameter(torch.rand(()))
                    self.loc_b = torch.nn.Parameter(torch.rand((3,)))

                def forward(self):
                    a = pyro.sample("a", dist.Normal(self.loc_a, 1))
                    with pyro.plate("data", 3, dim=-1):
                        b = pyro.sample("b", dist.Normal(self.loc_b, 1))
                        return {"a": a, "b": b}

            model = SimpleModel()
            guide = SimpleGuide()
            predictive = pyro.infer.Predictive(
                model, guide=guide, num_samples=10, return_sites=["y"]
            )
            points = predictive()
            influence = linearize(
                PredictiveModel(model, guide),
                num_samples_outer=1000,
                num_samples_inner=1000,
            )

            influence(points)

    .. note::

        * Since the efficient influence function is approximated using Monte Carlo, the result
          of this function is stochastic, i.e., evaluating this function on the same ``points``
          can result in different values. To reduce variance, increase ``num_samples_outer`` and
          ``num_samples_inner`` in ``linearize_kwargs``.

        * Currently, ``model`` cannot contain any ``pyro.param`` statements.
          This issue will be addressed in a future release:
          https://github.com/BasisResearch/chirho/issues/393.
    """
    if len(models) > 1:
        raise NotImplementedError("Only unary version of linearize is implemented.")
    else:
        (model,) = models

    assert isinstance(model, torch.nn.Module)
    if num_samples_inner is None:
        num_samples_inner = num_samples_outer**2

    predictive = pyro.infer.Predictive(
        model,
        num_samples=num_samples_outer,
        parallel=True,
    )

    batched_log_prob: BatchedNMCLogMarginalLikelihood[P, torch.Tensor] = (
        BatchedNMCLogMarginalLikelihood(
            model, num_samples=num_samples_inner, max_plate_nesting=max_plate_nesting
        )
    )
    log_prob_params, batched_func_log_prob = make_functional_call(batched_log_prob)
    log_prob_params_numel: int = sum(p.numel() for p in log_prob_params.values())
    if cg_iters is None:
        cg_iters = log_prob_params_numel
    else:
        cg_iters = min(cg_iters, log_prob_params_numel)
    cg_solver = functools.partial(
        conjugate_gradient_solve, cg_iters=cg_iters, residual_tol=residual_tol
    )

    def _fn(
        points: Point[T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ParamDict:
        with torch.no_grad():
            data: Point[T] = predictive(*args, **kwargs)
            data = {k: data[k] for k in points.keys()}
        fvp = make_empirical_fisher_vp(
            batched_func_log_prob, log_prob_params, data, *args, **kwargs
        )
        pinned_fvp = reset_rng_state(pyro.util.get_rng_state())(fvp)
        pinned_fvp_batched = torch.func.vmap(
            lambda v: pinned_fvp(v), randomness="different"
        )

        def bound_batched_func_log_prob(p: ParamDict) -> torch.Tensor:
            return batched_func_log_prob(p, points, *args, **kwargs)

        if pointwise_influence:
            score_fn = torch.func.jacrev(bound_batched_func_log_prob)
            point_scores = score_fn(log_prob_params)
        else:
            score_fn = torch.func.vjp(bound_batched_func_log_prob, log_prob_params)[1]
            N_pts = points[next(iter(points))].shape[0]  # type: ignore
            point_scores = score_fn(1 / N_pts * torch.ones(N_pts))[0]
            point_scores = {k: v.unsqueeze(0) for k, v in point_scores.items()}
        return cg_solver(pinned_fvp_batched, point_scores)

    return _fn



from typing import Optional

import pyro
import torch

from chirho.indexed.handlers import DependentMaskMessenger
from chirho.indexed.ops import IndexSet, get_index_plates, indexset_as_mask


def get_factual_indices() -> IndexSet:
    """
    Helpful operation used with :class:`MultiWorldCounterfactual`
    that returns an :class:`IndexSet` corresponding to the factual world,
    i.e. the world with index 0 for each index variable where no interventions
    have been performed.

    :return: IndexSet corresponding to the factual world.
    """
    return IndexSet(**{name: {0} for name in get_index_plates().keys()})


class SelectCounterfactual(DependentMaskMessenger):
    """
    Effect handler to include only log-density terms from counterfactual worlds.
    This implementation piggybacks on Pyro's existing masking functionality,
    as used in :class:`pyro.poutine.mask.MaskMessenger` and elsewhere.

    Useful for transformations that require different behavior in the factual
    and counterfactual worlds, such as conditioning.

    .. note:: Semantically equivalent to applying the following at each sample site:

        pyro.poutine.mask(mask=~indexset_as_mask(get_factual_indices()))
    """

    @staticmethod
    def get_mask(
        dist: pyro.distributions.Distribution,
        value: Optional[torch.Tensor],
        device: torch.device = torch.device("cpu"),
        name: Optional[str] = None,
    ) -> torch.Tensor:
        indices = get_factual_indices()
        return ~indexset_as_mask(indices, device=device)  # negate == complement


class SelectFactual(DependentMaskMessenger):
    """
    Effect handler to include only log-density terms from the factual world.
    This implementation piggybacks on Pyro's existing masking functionality,
    as used in :class:`pyro.poutine.mask.MaskMessenger` and elsewhere.

    Useful for transformations that require different behavior in the factual
    and counterfactual worlds, such as conditioning.

    .. note:: Semantically equivalent to applying the following at each sample site:

        pyro.poutine.mask(mask=indexset_as_mask(get_factual_indices()))
    """

    @staticmethod
    def get_mask(
        dist: pyro.distributions.Distribution,
        value: Optional[torch.Tensor] = None,
        device: torch.device = torch.device("cpu"),
        name: Optional[str] = None,
    ) -> torch.Tensor:
        indices = get_factual_indices()
        return indexset_as_mask(indices, device=device)



from __future__ import annotations

from typing import Any, Dict, TypeVar

import pyro

from chirho.counterfactual.handlers.ambiguity import FactualConditioningMessenger
from chirho.counterfactual.ops import split
from chirho.indexed.handlers import IndexPlatesMessenger
from chirho.indexed.ops import get_index_plates
from chirho.interventional.ops import intervene

T = TypeVar("T")


class BaseCounterfactualMessenger(FactualConditioningMessenger):
    """
    Base class for counterfactual handlers.

    :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` is an effect handler
    for imbuing :func:`~chirho.interventional.ops.intervene` operations with world-splitting
    semantics that is useful for downstream causal and counterfactual reasoning. Specifically,
    :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` handles
    :func:`~chirho.interventional.ops.intervene` by instantiating the primitive operation
    :func:`~chirho.counterfactual.ops.split`, which is then subsequently handled by subclasses
    such as :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual`.
    """

    @staticmethod
    def _pyro_intervene(msg: Dict[str, Any]) -> None:
        msg["stop"] = True
        if msg["args"][1] is not None:
            obs, acts = msg["args"][0], msg["args"][1]
            acts = acts(obs) if callable(acts) else acts
            acts = (acts,) if not isinstance(acts, tuple) else acts
            msg["value"] = split(obs, acts, name=msg["name"], **msg["kwargs"])
            msg["done"] = True

    @staticmethod
    def _pyro_preempt(msg: Dict[str, Any]) -> None:
        if msg["kwargs"].get("name", None) is None:
            msg["kwargs"]["name"] = msg["name"]


class SingleWorldCounterfactual(BaseCounterfactualMessenger):
    """
    Trivial counterfactual handler that returns the intervened value.

    :class:`~chirho.counterfactual.handlers.counterfactual.SingleWorldCounterfactual` is an effect handler
    that subclasses :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` and
    handles :func:`~chirho.counterfactual.ops.split` primitive operations. See the documentation for
    :func:`~chirho.counterfactual.ops.split` for more details about the interaction between the enclosing
    counterfactual handler and the induced joint marginal distribution over factual and counterfactual variables.

    :class:`~chirho.counterfactual.handlers.counterfactual.SingleWorldCounterfactual` handles
    :func:`~chirho.counterfactual.ops.split` by returning only the final element in the collection
    of intervention assignments ``acts``, ignoring all other intervention assignments and observed values ``obs``.
    This can be thought of as marginalizing out all of the factual and counterfactual variables except for the
    counterfactual induced by the final element in the collection of intervention assignments in the probabilistic
    program. ::

        >>> with SingleWorldCounterfactual():
        ...     x = torch.tensor(1.)
        ...     x = intervene(x, torch.tensor(0.))
        >>> assert (x == torch.tensor(0.))
    """

    @pyro.poutine.block(hide_types=["intervene"])
    def _pyro_split(self, msg: Dict[str, Any]) -> None:
        obs, acts = msg["args"]
        msg["value"] = intervene(obs, acts[-1], **msg["kwargs"])
        msg["done"] = True
        msg["stop"] = True


class SingleWorldFactual(BaseCounterfactualMessenger):
    """
    Trivial counterfactual handler that returns the observed value.

    :class:`~chirho.counterfactual.handlers.counterfactual.SingleWorldFactual` is an effect handler
    that subclasses :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` and
    handles :func:`~chirho.counterfactual.ops.split` primitive operations. See the documentation for
    :func:`~chirho.counterfactual.ops.split` for more details about the interaction between the enclosing
    counterfactual handler and the induced joint marginal distribution over factual and counterfactual variables.

    :class:`~chirho.counterfactual.handlers.counterfactual.SingleWorldFactual` handles
    :func:`~chirho.counterfactual.ops.split` by returning only the observed value ``obs``,
    ignoring all intervention assignments ``act``. This can be thought of as marginalizing out
    all of the counterfactual variables in the probabilistic program. ::

        >>> with SingleWorldFactual():
        ...    x = torch.tensor(1.)
        ...    x = intervene(x, torch.tensor(0.))
        >>> assert (x == torch.tensor(1.))
    """

    @staticmethod
    def _pyro_split(msg: Dict[str, Any]) -> None:
        obs, _ = msg["args"]
        msg["value"] = obs
        msg["done"] = True
        msg["stop"] = True


class MultiWorldCounterfactual(IndexPlatesMessenger, BaseCounterfactualMessenger):
    """
    Counterfactual handler that returns all observed and intervened values.

    :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual` is an effect handler
    that subclasses :class:`~chirho.indexed.handlers.IndexPlatesMessenger` and
    :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` base classes.


    .. note:: Handlers that subclass :class:`~chirho.indexed.handlers.IndexPlatesMessenger` such as
       :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual` return tensors that can
       be cumbersome to index into directly. Therefore, we strongly recommend using ``chirho``'s indexing operations
       :func:`~chirho.indexed.ops.gather` and :class:`~chirho.indexed.ops.IndexSet` whenever using
       :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual` handlers.

    :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual`
    handles :func:`~chirho.counterfactual.ops.split` primitive operations. See the documentation for
    :func:`~chirho.counterfactual.ops.split` for more details about the interaction between the enclosing
    counterfactual handler and the induced joint marginal distribution over factual and counterfactual variables.

    :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual` handles
    :func:`~chirho.counterfactual.ops.split` by returning all observed values ``obs`` and intervened values ``act``.
    This can be thought of as returning the full joint distribution over all factual and counterfactual variables. ::

        >>> with MultiWorldCounterfactual():
        ...    x = torch.tensor(1.)
        ...    x = intervene(x, torch.tensor(0.), name="x_ax_1")
        ...    x = intervene(x, torch.tensor(2.), name="x_ax_2")
        ...    x_factual = gather(x, IndexSet(x_ax_1={0}, x_ax_2={0}))
        ...    x_counterfactual_1 = gather(x, IndexSet(x_ax_1={1}, x_ax_2={0}))
        ...    x_counterfactual_2 = gather(x, IndexSet(x_ax_1={0}, x_ax_2={1}))

        >>> assert(x_factual.squeeze() == torch.tensor(1.))
        >>> assert(x_counterfactual_1.squeeze() == torch.tensor(0.))
        >>> assert(x_counterfactual_2.squeeze() == torch.tensor(2.))
    """

    fresh_prefix: str = "__fresh_split__"

    @classmethod
    def _pyro_split(cls, msg: Dict[str, Any]) -> None:
        if msg["name"] is None:
            index_plates = get_index_plates()
            name, fresh_suffix = cls.fresh_prefix, len(index_plates)
            while name in index_plates:
                name = f"{cls.fresh_prefix}{fresh_suffix}"
                fresh_suffix += 1
        else:
            name = msg["name"]
        msg["kwargs"]["name"] = msg["name"] = name


class TwinWorldCounterfactual(IndexPlatesMessenger, BaseCounterfactualMessenger):
    """
    Counterfactual handler that returns all observed values and the final intervened value.

    :class:`~chirho.counterfactual.handlers.counterfactual.TwinWorldCounterfactual` is an effect handler
    that subclasses :class:`~chirho.indexed.handlers.IndexPlatesMessenger` and
    :class:`~chirho.counterfactual.handlers.counterfactual.BaseCounterfactualMessenger` base classes.


    .. note:: Handlers that subclass :class:`~chirho.indexed.handlers.IndexPlatesMessenger` such as
       :class:`~chirho.counterfactual.handlers.counterfactual.TwinWorldCounterfactual` return tensors that can
       be cumbersome to index into directly. Therefore, we strongly recommend using ``chirho``'s indexing operations
       :func:`~chirho.indexed.ops.gather` and :class:`~chirho.indexed.ops.IndexSet` whenever using
       :class:`~chirho.counterfactual.handlers.counterfactual.TwinWorldCounterfactual` handlers.

    :class:`~chirho.counterfactual.handlers.counterfactual.TwinWorldCounterfactual`
    handles :func:`~chirho.counterfactual.ops.split` primitive operations. See the documentation for
    :func:`~chirho.counterfactual.ops.split` for more details about the interaction between the enclosing
    counterfactual handler and the induced joint marginal distribution over factual and counterfactual variables.

    :class:`~chirho.counterfactual.handlers.counterfactual.TwinWorldCounterfactual` handles
    :func:`~chirho.counterfactual.ops.split` by returning the observed values ``obs`` and the
    final intervened values ``act`` in the probabilistic program. This can be thought of as returning
    the joint distribution over factual and counterfactual variables, marginalizing out all but the final
    configuration of intervention assignments in the probabilistic program. ::

        >>> with TwinWorldCounterfactual():
        ...    x = torch.tensor(1.)
        ...    x = intervene(x, torch.tensor(0.))
        ...    x = intervene(x, torch.tensor(2.))
        >>> # TwinWorldCounterfactual ignores the first intervention
        >>> assert(x.squeeze().shape == torch.Size([2]))
        >>> assert(x.squeeze()[0] == torch.tensor(1.))
        >>> assert(x.squeeze()[1] == torch.tensor(2.))
    """

    fresh_prefix: str = "__fresh_split__"

    @classmethod
    def _pyro_split(cls, msg: Dict[str, Any]) -> None:
        msg["kwargs"]["name"] = msg["name"] = cls.fresh_prefix



from __future__ import annotations

import functools
from typing import Tuple, TypeVar

import pyro

from chirho.indexed.ops import IndexSet, gather, scatter_n
from chirho.interventional.ops import Intervention, intervene

S = TypeVar("S")
T = TypeVar("T")


@pyro.poutine.runtime.effectful(type="split")
@functools.partial(pyro.poutine.block, hide_types=["intervene"])
def split(obs: T, acts: Tuple[Intervention[T], ...], **kwargs) -> T:
    """
    Effectful primitive operation for "splitting" a combination of observational and interventional values in a
    probabilistic program into counterfactual worlds.

    :func:`~chirho.counterfactual.ops.split` returns the result of the effectful primitive operation
    :func:`~chirho.indexed.ops.scatter` applied to the concatenation of the ``obs`` and ``acts`` arguments,
    where ``obs`` represents the single observed value in the probabilistic program and ``acts`` represents
    the collection of intervention assignments.

    In a probabilistic program, :func:`split` induces a joint distribution over factual and counterfactual variables,
    where some variables are implicitly marginalized out by enclosing counterfactual handlers. For example,
    :func:`split` in the context of a :class:`~chirho.counterfactual.handlers.counterfactual.MultiWorldCounterfactual`
    handler induces a joint distribution over all combinations of ``obs`` and ``acts``, whereas
    :class:`~chirho.counterfactual.handlers.counterfactual.SingleWorldFactual` marginalizes out all ``acts``.

    :param obs: The observed value.
    :param acts: The interventions to apply.

    """
    name = kwargs.get("name", None)
    obs_idx = IndexSet(**{name: {0}})
    act_values = {obs_idx: gather(obs, obs_idx, **kwargs)}
    for i, act in enumerate(acts):
        act_idx = IndexSet(**{name: {i + 1}})
        act_values[act_idx] = gather(
            intervene(act_values[obs_idx], act, **kwargs), act_idx, **kwargs
        )

    return scatter_n(act_values, event_dim=kwargs.get("event_dim", 0))



from typing import Generic, Mapping, TypeVar

import pyro
import torch

from chirho.explainable.ops import preempt
from chirho.interventional.ops import Intervention

S = TypeVar("S")
T = TypeVar("T")


class Preemptions(Generic[T], pyro.poutine.messenger.Messenger):
    """
    Effect handler that applies the operation :func:`~chirho.explainable.ops.preempt`
    to sample sites in a probabilistic program,
    similar to the handler :func:`~chirho.observational.handlers.condition`
    for :func:`~chirho.observational.ops.observe` .
    or the handler :func:`~chirho.interventional.handlers.do`
    for :func:`~chirho.interventional.ops.intervene` .

    See the documentation for :func:`~chirho.explainable.ops.preempt` for more details.

    This handler introduces an auxiliary discrete random variable at each preempted sample site
    whose name is the name of the sample site prefixed by ``prefix``, and
    whose value is used as the ``case`` argument to :func:`preempt`,
    to determine whether the preemption returns the present value of the site
    or the new value specified for the site in ``actions``

    The distributions of the auxiliary discrete random variables are parameterized by ``bias``.
    By default, ``bias == 0`` and the value returned by the sample site is equally likely
    to be the factual case (i.e. the present value of the site) or one of the counterfactual cases
    (i.e. the new value(s) specified for the site in ``actions``).
    When ``0 < bias <= 0.5``, the preemption is less than equally likely to occur.
    When ``-0.5 <= bias < 0``, the preemption is more than equally likely to occur.

    More specifically, the probability of the factual case is ``0.5 - bias``,
    and the probability of each counterfactual case is ``(0.5 + bias) / num_actions``,
    where ``num_actions`` is the number of counterfactual actions for the sample site (usually 1).

    :param actions: A mapping from sample site names to interventions.
    :param bias: The scalar bias towards not intervening. Must be between -0.5 and 0.5.
    :param prefix: The prefix for naming the auxiliary discrete random variables.
    """

    actions: Mapping[str, Intervention[T]]
    prefix: str
    bias: float

    def __init__(
        self,
        actions: Mapping[str, Intervention[T]],
        *,
        prefix: str = "__witness_split_",
        bias: float = 0.0,
    ):
        assert -0.5 <= bias <= 0.5, "bias must be between -0.5 and 0.5"
        self.actions = actions
        self.bias = bias
        self.prefix = prefix
        super().__init__()

    def _pyro_post_sample(self, msg):
        try:
            action = self.actions[msg["name"]]
        except KeyError:
            return

        action = (action,) if not isinstance(action, tuple) else action
        num_actions = len(action) if isinstance(action, tuple) else 1
        weights = torch.tensor(
            [0.5 - self.bias] + ([(0.5 + self.bias) / num_actions] * num_actions),
            device=msg["value"].device,
        )
        case_dist = pyro.distributions.Categorical(probs=weights)
        case = pyro.sample(f"{self.prefix}{msg['name']}", case_dist)

        msg["value"] = preempt(
            msg["value"],
            action,
            case,
            event_dim=len(msg["fn"].event_shape),
            name=f"{self.prefix}{msg['name']}",
        )



import collections
import functools
from typing import Any, Callable, Dict, List, Optional

import pyro
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin
from typing_extensions import ParamSpec

from chirho.indexed.internals import (
    _LazyPlateMessenger,
    add_indices,
    get_sample_msg_device,
)
from chirho.indexed.ops import union

P = ParamSpec("P")


class IndexPlatesMessenger(pyro.poutine.messenger.Messenger):
    plates: Dict[str, pyro.poutine.indep_messenger.IndepMessenger]
    first_available_dim: int

    def __init__(self, first_available_dim: Optional[int] = None):
        if first_available_dim is None:
            first_available_dim = -5  # conservative default for 99% of models
        assert first_available_dim < 0
        self._orig_dim = first_available_dim
        self.first_available_dim = first_available_dim
        self.plates = collections.OrderedDict()
        super().__init__()

    def __enter__(self):
        ret = super().__enter__()
        for name in self.plates.keys():
            self.plates[name].__enter__()
        return ret

    def __exit__(self, exc_type, exc_value, traceback):
        for name in reversed(list(self.plates.keys())):
            self.plates[name].__exit__(exc_type, exc_value, traceback)
        return super().__exit__(exc_type, exc_value, traceback)

    def __call__(self, fn: Callable) -> Callable:
        handled_fn = super().__call__(fn)

        # IndexPlatesMessenger is a stateful handler, and by default
        # does not clear its state after exiting a context to support REPL usage.
        # This wrapper ensures that state is cleared after exiting a context
        # when IndexPlatesMessenger is used as a decorator.
        @functools.wraps(handled_fn)
        def wrapped_handled_fn(*args, **kwargs):
            try:
                return handled_fn(*args, **kwargs)
            finally:
                if self not in pyro.poutine.runtime._PYRO_STACK:
                    for p in list(self.plates):
                        assert p not in pyro.poutine.runtime._PYRO_STACK
                        del self.plates[p]
                    self.first_available_dim = self._orig_dim

        return wrapped_handled_fn

    def _pyro_get_index_plates(self, msg):
        msg["value"] = {name: plate.frame for name, plate in self.plates.items()}
        msg["done"], msg["stop"] = True, True

    def _enter_index_plate(self, plate: _LazyPlateMessenger) -> _LazyPlateMessenger:
        try:
            plate.__enter__()
        except ValueError as e:
            if "collide at dim" in str(e):
                raise ValueError(
                    f"{self} was unable to allocate an index plate dimension "
                    f"at dimension {self.first_available_dim}.\n"
                    f"Try setting a value less than {self._orig_dim} for `first_available_dim` "
                    "that is less than the leftmost (most negative) plate dimension in your model."
                )
            else:
                raise e
        stack: List[pyro.poutine.messenger.Messenger] = pyro.poutine.runtime._PYRO_STACK
        stack.pop(stack.index(plate))
        stack.insert(stack.index(self) + len(self.plates) + 1, plate)
        return plate

    def _pyro_add_indices(self, msg):
        (indexset,) = msg["args"]
        for name, indices in indexset.items():
            if name not in self.plates:
                new_size = max(max(indices) + 1, len(indices))
                # Push the new plate onto Pyro's handler stack at a location
                # adjacent to this IndexPlatesMessenger instance so that
                # any handlers pushed after this IndexPlatesMessenger instance
                # are still guaranteed to exit safely in the correct order.
                self.plates[name] = self._enter_index_plate(
                    _LazyPlateMessenger(
                        name=name,
                        dim=self.first_available_dim,
                        size=new_size,
                    )
                )
                self.first_available_dim -= 1
            else:
                assert (
                    0
                    <= min(indices)
                    <= len(indices) - 1
                    <= max(indices)
                    < self.plates[name].size
                ), f"cannot add {name}={indices} to {self.plates[name].size}"

    def _pyro_scatter_n(self, msg: Dict[str, Any]) -> None:
        add_indices(union(*msg["args"][0].keys()))
        msg["stop"] = True


class DependentMaskMessenger(pyro.poutine.messenger.Messenger):
    """
    Abstract base class for effect handlers that select a subset of worlds.
    """

    def get_mask(
        self,
        dist: pyro.distributions.Distribution,
        value: Optional[torch.Tensor],
        device: torch.device = torch.device("cpu"),
        name: Optional[str] = None,
    ) -> torch.Tensor:
        raise NotImplementedError

    def _pyro_sample(self, msg: pyro.poutine.messenger.Message) -> None:
        if pyro.poutine.util.site_is_subsample(msg):
            return

        assert isinstance(msg["fn"], TorchDistributionMixin)

        device = get_sample_msg_device(msg["fn"], msg["value"])
        name = msg["name"] if "name" in msg else None
        mask = self.get_mask(msg["fn"], msg["value"], device=device, name=name)
        msg["mask"] = mask if msg["mask"] is None else msg["mask"] & mask

        # expand distribution to make sure two copies of a variable are sampled
        msg["fn"] = msg["fn"].expand(
            torch.broadcast_shapes(msg["fn"].batch_shape, mask.shape)
        )


@pyro.poutine.block()
@pyro.validation_enabled(False)
@torch.no_grad()
def guess_max_plate_nesting(
    model: Callable[P, Any], guide: Callable[P, Any], *args: P.args, **kwargs: P.kwargs
) -> int:
    """
    Guesses the maximum plate nesting level by running `pyro.infer.Trace_ELBO`

    :param model: Python callable containing Pyro primitives.
    :type model: Callable[P, Any]
    :param guide: Python callable containing Pyro primitives.
    :type guide: Callable[P, Any]
    :return: maximum plate nesting level
    :rtype: int
    """
    elbo = pyro.infer.Trace_ELBO()
    elbo._guess_max_plate_nesting(model, guide, args, kwargs)
    return elbo.max_plate_nesting



from typing import Callable, Iterable, MutableMapping, Optional, TypeVar

import pyro
import pyro.distributions.constraints as constraints
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin

from chirho.counterfactual.handlers.selection import get_factual_indices
from chirho.explainable.internals import uniform_proposal
from chirho.indexed.ops import IndexSet, gather, indices_of, scatter_n
from chirho.observational.handlers import soft_eq, soft_neq
from chirho.observational.ops import Observation

S = TypeVar("S")
T = TypeVar("T")


def sufficiency_intervention(
    support: constraints.Constraint,
    antecedents: Iterable[str] = [],
    sufficiency_world=2,
) -> Callable[[T], T]:
    """
    Creates a sufficiency intervention for a single sample site, determined by
    the site name, intervening to keep the value as in the factual world with
    respect to the antecedents.

    :param support: The support constraint for the site.
    :param name: The sample site name.

    :return: A function that takes a `torch.Tensor` as input
        and returns the factual value at the named site as a tensor.

    Example::

        >>> with MultiWorldCounterfactual() as mwc:
        >>>     value = pyro.sample("value", proposal_dist)
        >>>     intervention = sufficiency_intervention(support)
        >>>     value = intervene(value, intervention)
    """

    def _sufficiency_intervention(value: T) -> T:

        indices = IndexSet(
            **{
                name: sufficiency_world
                for name, ind in get_factual_indices().items()
                if name in antecedents
            }
        )

        sufficiency_value = gather(
            value,
            indices,
            event_dim=support.event_dim,
        )
        return sufficiency_value

    return _sufficiency_intervention


def random_intervention(
    support: constraints.Constraint,
    name: str,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Creates a random-valued intervention for a single sample site, determined by
    by the distribution support, and site name.

    :param support: The support constraint for the sample site.
    :param name: The name of the auxiliary sample site.

    :return: A function that takes a `torch.Tensor` as input
        and returns a random sample over the pre-specified support of the same
        event shape as the input tensor.

    Example::

        >>> support = pyro.distributions.constraints.real
        >>> intervention_fn = random_intervention(support, name="random_value")
        >>> with chirho.interventional.handlers.do(actions={"x": intervention_fn}):
        ...   x = pyro.deterministic("x", torch.tensor(2.))
        >>> assert x != 2
    """

    def _random_intervention(value: torch.Tensor) -> torch.Tensor:

        event_shape = value.shape[len(value.shape) - support.event_dim :]  # type: ignore

        proposal_dist = uniform_proposal(support, event_shape=event_shape)
        return pyro.sample(name, proposal_dist)

    return _random_intervention


def undo_split(
    support: constraints.Constraint, antecedents: Iterable[str] = []
) -> Callable[[T], T]:
    """
    A helper function that undoes an upstream :func:`~chirho.counterfactual.ops.split` operation,
    meant to be used to create arguments to pass to :func:`~chirho.interventional.ops.intervene` ,
    :func:`~chirho.counterfactual.ops.split`  or :func:`~chirho.explainable.ops.preempt`.
    Works by gathering the factual value and scattering it back into two alternative cases.

    :param support: The support constraint for the site at which :func:`split` is being undone.
    :param antecedents: A list of upstream intervened sites which induced the :func:`split`
        to be reversed.
    :return: A callable that applied to a site value object returns a site value object in which
        the factual value has been scattered back into two alternative cases.
    """

    def _undo_split(value: T) -> T:
        antecedents_ = {
            a: v
            for a, v in indices_of(value, event_dim=support.event_dim).items()
            if a in antecedents
        }

        factual_value = gather(
            value,
            IndexSet(**{antecedent: {0} for antecedent in antecedents_.keys()}),
            event_dim=support.event_dim,
        )

        # TODO exponential in len(antecedents) - add an indexed.ops.expand to do this cheaply
        index_keys: Iterable[MutableMapping[str, Iterable[int]]] = list()
        for a, v in antecedents_.items():
            if index_keys == []:
                index_keys = [dict({a: {value}}.items()) for value in v]
            else:
                temp_index_keys = []
                for i in index_keys:
                    temp_index_keys.extend(
                        [
                            dict(tuple(dict(i).items()) + tuple({a: {value}}.items()))
                            for value in v
                        ]
                    )
                index_keys = temp_index_keys
        index_keys = index_keys if index_keys != [] else [{}]

        return scatter_n(
            {IndexSet(**ind_key): factual_value for ind_key in index_keys},
            event_dim=support.event_dim,
        )

    return _undo_split


def consequent_eq(
    support: constraints.Constraint,
    antecedents: Iterable[str] = [],
    **kwargs,
) -> Callable[[T], torch.Tensor]:
    """
    A helper function for assessing whether values at a site are close to their observed values, assigning
    a small negative value close to zero if a value is close to its observed state and a large negative value otherwise.

    :param support: The support constraint for the consequent site.
    :param antecedents: A list of names of upstream intervened sites to consider when assessing similarity.

    :return: A callable which applied to a site value object (``consequent``), returns a tensor where each
             element indicates the extent to which the corresponding element of ``consequent``
             is close to its factual value.
    """

    def _consequent_eq(consequent: T) -> torch.Tensor:
        indices = IndexSet(
            **{
                name: ind
                for name, ind in get_factual_indices().items()
                if name in antecedents
            }
        )
        eq = soft_eq(
            support,
            consequent,
            gather(consequent, indices, event_dim=support.event_dim),
            **kwargs,
        )
        return eq

    return _consequent_eq


def consequent_neq(
    support: constraints.Constraint,
    antecedents: Iterable[str] = [],
    **kwargs,
) -> Callable[[T], torch.Tensor]:
    """
    A helper function for assessing whether values at a site differ from their observed values, assigning
    a small negative value close to zero if a value differs from its observed state
    and a large negative value otherwise.

    :param support: The support constraint for the consequent site.
    :param antecedents: A list of names of upstream intervened sites to consider when assessing differences.

    :return: A callable which applied to a site value object (``consequent``), returns a tensor where each
             element indicates whether the corresponding element of ``consequent`` differs from its factual value.
    """

    def _consequent_neq(consequent: T) -> torch.Tensor:
        indices = IndexSet(
            **{
                name: ind
                for name, ind in get_factual_indices().items()
                if name in antecedents
            }
        )
        diff = soft_neq(
            support,
            consequent,
            gather(consequent, indices, event_dim=support.event_dim),
            **kwargs,
        )
        return diff

    return _consequent_neq


def consequent_eq_neq(
    support: constraints.Constraint,
    proposed_consequent: Optional[Observation[T]],
    antecedents: Iterable[str] = [],
    **kwargs,
) -> Callable[[T], torch.Tensor]:
    """
    A helper function for obtaining joint log prob of necessity and sufficiency. Assumes that
    the necessity intervention has been applied in counterfactual world 1 and sufficiency intervention in
    counterfactual world 2 (these can be passed as kwargs).

    :param support: The support constraint for the consequent site.
    :param antecedents: A list of names of upstream intervened sites to consider when composing the joint log prob.

    :return: A callable which applied to a site value object (``consequent``), returns a tensor with log prob sums
    of values resulting from necessity and sufficiency interventions, in appropriate counterfactual worlds.
    """

    def _consequent_eq_neq(consequent: T) -> torch.Tensor:
        necessity_world = kwargs.get("necessity_world", 1)
        sufficiency_world = kwargs.get("sufficiency_world", 2)

        necessity_indices = IndexSet(
            **{
                name: {necessity_world}
                for name in indices_of(consequent, event_dim=support.event_dim).keys()
                if name in antecedents
            }
        )
        sufficiency_indices = IndexSet(
            **{
                name: {sufficiency_world}
                for name in indices_of(consequent, event_dim=support.event_dim).keys()
                if name in antecedents
            }
        )

        necessity_value = gather(
            consequent, necessity_indices, event_dim=support.event_dim
        )
        sufficiency_value = gather(
            consequent, sufficiency_indices, event_dim=support.event_dim
        )

        necessity_log_probs = (
            soft_neq(
                support,
                necessity_value,
                proposed_consequent,
                **kwargs,
            )
            if proposed_consequent is not None
            else soft_neq(
                support,
                necessity_value,
                sufficiency_value,
                **kwargs,
            )
        )
        sufficiency_log_probs = (
            soft_eq(support, sufficiency_value, proposed_consequent, **kwargs)
            if proposed_consequent is not None
            else torch.zeros_like(necessity_log_probs)
        )

        FACTUAL_NEC_SUFF = torch.zeros_like(sufficiency_log_probs)

        index_keys = set(antecedents)
        null_index = IndexSet(**{name: {0} for name in index_keys})

        nec_suff_log_probs_partitioned = {
            **{null_index: FACTUAL_NEC_SUFF},
            **{
                IndexSet(**{antecedent: {ind} for antecedent in index_keys}): log_prob
                for ind, log_prob in zip(
                    [necessity_world, sufficiency_world],
                    [necessity_log_probs, sufficiency_log_probs],
                )
            },
        }

        new_value = scatter_n(
            nec_suff_log_probs_partitioned,
            event_dim=0,
        )
        return new_value

    return _consequent_eq_neq


class ExtractSupports(pyro.poutine.messenger.Messenger):
    """
    A Pyro Messenger for inferring distribution constraints.

    :return: An instance of ``ExtractSupports`` with a new attribute: ``supports``,
            a dictionary mapping variable names to constraints for all variables in the model.

    Example:

        >>> def mixed_supports_model():
        >>>     uniform_var = pyro.sample("uniform_var", dist.Uniform(1, 10))
        >>>     normal_var = pyro.sample("normal_var", dist.Normal(3, 15))
        >>> with ExtractSupports() as s:
        ...      mixed_supports_model()
        >>> print(s.supports)
    """

    supports: MutableMapping[str, pyro.distributions.constraints.Constraint]

    def __init__(self):
        super(ExtractSupports, self).__init__()

        self.supports = {}

    def _pyro_post_sample(self, msg: pyro.poutine.messenger.Message) -> None:
        if not pyro.poutine.util.site_is_subsample(msg):
            assert msg["name"] is not None
            assert isinstance(msg["fn"], TorchDistributionMixin)
            self.supports[msg["name"]] = msg["fn"].support



from typing import Any, Callable, Generic, Mapping, Optional, TypeVar, cast

import pyro
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin
from typing_extensions import ParamSpec

from chirho.indexed.handlers import IndexPlatesMessenger
from chirho.indexed.ops import indices_of
from chirho.observational.handlers.condition import Observations
from chirho.observational.internals import (
    bind_leftmost_dim,
    site_is_delta,
    unbind_leftmost_dim,
)
from chirho.observational.ops import Observation

P = ParamSpec("P")
S = TypeVar("S")
T = TypeVar("T")

Point = Mapping[str, Observation[T]]


class BatchedLatents(pyro.poutine.messenger.Messenger):
    """
    Effect handler that adds a fresh batch dimension to all latent ``sample`` sites.
    Similar to wrapping a Pyro model in a ``pyro.plate`` context, but uses the machinery
    in ``chirho.indexed`` to automatically allocate and track the fresh batch dimension
    based on the ``name`` argument to ``BatchedLatents`` .

    .. warning:: Must be used in conjunction with :class:`~chirho.indexed.handlers.IndexPlatesMessenger` .

    :param int num_particles: Number of particles to use for parallelization.
    :param str name: Name of the fresh batch dimension.
    """

    num_particles: int
    name: str

    def __init__(self, num_particles: int, *, name: str = "__particles_mc"):
        assert num_particles > 0
        assert len(name) > 0
        self.num_particles = num_particles
        self.name = name
        super().__init__()

    def _pyro_sample(self, msg: pyro.poutine.messenger.Message) -> None:
        if (
            self.num_particles > 1
            and msg["value"] is None
            and not pyro.poutine.util.site_is_factor(msg)
            and not pyro.poutine.util.site_is_subsample(msg)
            and not site_is_delta(msg)
            and self.name not in indices_of(msg["fn"])
        ):
            assert isinstance(msg["fn"], TorchDistributionMixin)
            msg["fn"] = unbind_leftmost_dim(
                msg["fn"].expand((1,) + msg["fn"].batch_shape),
                self.name,
                size=self.num_particles,
            )


class BatchedObservations(Generic[T], Observations[T]):
    """
    Effect handler that takes a dictionary of observation values for ``sample`` sites
    that are assumed to be batched along their leftmost dimension, adds a fresh named
    dimension using the machinery in ``chirho.indexed``, and reshapes the observation
    values so that the new ``chirho.observational.observe`` sites are batched along
    the fresh named dimension.

    Useful in combination with ``pyro.infer.Predictive`` which returns a dictionary
    of values whose leftmost dimension is a batch dimension over independent samples.

    .. warning:: Must be used in conjunction with :class:`~chirho.indexed.handlers.IndexPlatesMessenger` .

    :param Point[T] data: Dictionary of observation values.
    :param str name: Name of the fresh batch dimension.
    """

    name: str

    def __init__(self, data: Point[T], *, name: str = "__particles_data"):
        assert len(name) > 0
        self.name = name
        super().__init__(data)

    def _pyro_observe(self, msg: dict) -> None:
        super()._pyro_observe(msg)
        if msg["kwargs"]["name"] in self.data:
            rv, obs = msg["args"]
            event_dim = (
                len(rv.event_shape)
                if hasattr(rv, "event_shape")
                else msg["kwargs"].get("event_dim", 0)
            )
            batch_obs = unbind_leftmost_dim(obs, self.name, event_dim=event_dim)
            msg["args"] = (rv, batch_obs)


class InferDictPredictive(pyro.poutine.runtime.InferDict):
    _model_predictive_site: bool


def _predictive_site_config(model_predictive_site: bool) -> InferDictPredictive:
    return {"_model_predictive_site": model_predictive_site}


class PredictiveModel(Generic[P, T], torch.nn.Module):
    """
    Given a Pyro model and guide, constructs a new model that behaves as if
    the latent ``sample`` sites in the original model (i.e. the prior)
    were replaced by their counterparts in the guide (i.e. the posterior).

    .. note:: Sites that only appear in the model are annotated in traces
        produced by the predictive model with ``infer={"_model_predictive_site": True}`` .

    :param model: Pyro model.
    :param guide: Pyro guide.
    """

    model: Callable[P, T]
    guide: Optional[Callable[P, Any]]

    def __init__(
        self,
        model: Callable[P, T],
        guide: Optional[Callable[P, Any]] = None,
    ):
        super().__init__()
        self.model = model
        self.guide = guide

    def forward(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Returns a sample from the posterior predictive distribution.

        :return: Sample from the posterior predictive distribution.
        :rtype: T
        """
        with pyro.poutine.infer_config(
            config_fn=lambda msg: _predictive_site_config(False)
        ):
            with pyro.poutine.trace() as guide_tr:
                if self.guide is not None:
                    self.guide(*args, **kwargs)

        block_guide_sample_sites = pyro.poutine.block(
            hide=[
                name
                for name, node in guide_tr.trace.nodes.items()
                if node["type"] == "sample"
            ]
        )

        with pyro.poutine.infer_config(
            config_fn=lambda msg: _predictive_site_config(True)
        ):
            with block_guide_sample_sites:
                with pyro.poutine.replay(trace=guide_tr.trace):
                    return self.model(*args, **kwargs)


class PredictiveFunctional(Generic[P, T], torch.nn.Module):
    """
    Functional that returns a batch of samples from the predictive
    distribution of a Pyro model. As with ``pyro.infer.Predictive`` ,
    the returned values are batched along their leftmost positional dimension.

    Similar to ``pyro.infer.Predictive(model, guide, num_samples, parallel=True)``
    when :class:`~chirho.observational.handlers.predictive.PredictiveModel` is used to construct
    the ``model`` argument and infer the ``sample`` sites whose values should be returned,
    and uses :class:`~BatchedLatents` to parallelize over samples from the model.

    .. warning:: ``PredictiveFunctional`` currently applies its own internal instance of
        :class:`~chirho.indexed.handlers.IndexPlatesMessenger` ,
        so it may not behave as expected if used within another enclosing
        :class:`~chirho.indexed.handlers.IndexPlatesMessenger` context.

    :param model: Pyro model.
    :param num_samples: Number of samples to return.
    """

    model: Callable[P, Any]
    num_samples: int

    def __init__(
        self,
        model: torch.nn.Module,
        *,
        num_samples: int = 1,
        max_plate_nesting: Optional[int] = None,
        name: str = "__particles_predictive",
    ):
        super().__init__()
        self.model = model
        self.num_samples = num_samples
        self._first_available_dim = (
            -max_plate_nesting - 1 if max_plate_nesting is not None else None
        )
        self._mc_plate_name = name

    def forward(self, *args: P.args, **kwargs: P.kwargs) -> Point[T]:
        """
        Returns a batch of samples from the posterior predictive distribution.

        :return: Dictionary of samples from the posterior predictive distribution.
        :rtype: Point[T]
        """
        with IndexPlatesMessenger(first_available_dim=self._first_available_dim):
            with pyro.poutine.trace() as model_tr:
                with BatchedLatents(self.num_samples, name=self._mc_plate_name):
                    with pyro.poutine.infer_config(
                        config_fn=lambda msg: _predictive_site_config(
                            cast(dict, msg.get("infer", {})).get(
                                "_model_predictive_site", False
                            )
                        )
                    ):
                        self.model(*args, **kwargs)

            result = {}
            for name, node in model_tr.trace.nodes.items():
                infer = cast(dict, node.get("infer", {}))
                if (
                    node["type"] == "sample"
                    and not pyro.poutine.util.site_is_subsample(node)
                    and infer.get("_model_predictive_site", False)
                ):
                    assert isinstance(node["fn"], TorchDistributionMixin)
                    result[name] = bind_leftmost_dim(
                        node["value"],
                        self._mc_plate_name,
                        event_dim=len(node["fn"].event_shape),
                    )
            return result



from __future__ import annotations

import numbers
from typing import Callable, Mapping, Optional, Tuple, TypeVar, Union

import pyro
import torch

R = Union[numbers.Real, torch.Tensor]
S = TypeVar("S")
T = TypeVar("T")

State = Mapping[str, T]
Dynamics = Callable[[State[T]], State[T]]


@pyro.poutine.runtime.effectful(type="simulate")
def simulate(
    dynamics: Dynamics[T],
    initial_state: State[T],
    start_time: R,
    end_time: R,
    **kwargs,
) -> State[T]:
    """
    Simulate a dynamical system for (``end_time - start_time``) units of time, starting the system at ``initial_state``,
    and rolling out the system according to the ``dynamics`` function. Note that this function is effectful, and must
    be within the context of a solver backend, such as :class:`~chirho.dynamical.handlers.solver.TorchDiffEq`.

    :param dynamics: A function that takes a state and returns the derivative of the state with respect to time.
    :type dynamics: :ref:`Dynamics[T] <type-alias-Dynamics>`

    :param initial_state: The initial state of the system.
    :type initial_state: :ref:`State[T] <type-alias-State>`

    :param start_time: The starting time of the simulation — a scalar.
    :type start_time: :ref:`R <type-alias-R>`

    :param end_time: The ending time of the simulation — a scalar.
    :type end_time: :ref:`R <type-alias-R>`

    :param kwargs: Additional keyword arguments to pass to the solver.

    :return: The final state of the system after the simulation.
    :rtype: :ref:`State[T] <type-alias-State>`
    """
    from chirho.dynamical.internals.solver import check_dynamics, simulate_point

    if pyro.settings.get("validate_dynamics"):
        check_dynamics(dynamics, initial_state, start_time, end_time, **kwargs)
    return simulate_point(dynamics, initial_state, start_time, end_time, **kwargs)


def on(
    predicate: Callable[[State[T]], bool],
    callback: Optional[
        Callable[[Dynamics[T], State[T]], Tuple[Dynamics[T], State[T]]]
    ] = None,
):
    """
    Creates a context manager that, when active, interrupts the first :func:`~chirho.dynamical.ops.simulate`
    call the first time that the ``predicate`` function applied to the current state returns ``True``.
    The ``callback`` function is then called with the current dynamics and state, and the return values
    are used as the new dynamics and state for the remainder of the simulation time.

    ``callback`` functions may invoke effectful operations such as :func:`~chirho.interventional.ops.intervene`
    that are then handled by the effect handlers around the :func:`~chirho.dynamical.ops.simulate` call.

    ``on`` may be used with two arguments to immediately create a context manager or higher-order function,
    or invoked with one ``predicate`` argument as a decorator for creating a context manager
    or higher-order function from a ``callback`` functions::

        >>> @on(lambda state: state["x"] > 0)
        ... def intervene_on_positive_x(dynamics, state):
        ...     return dynamics, intervene(state, {"x": state["x"] - 100})
        ...
        >>> with solver:
        ...     with intervene_on_positive_x:
        ...         xf = simulate(dynamics, {"x": 0}, 0, 1)["x"]
        ...
        >>> assert xf < 0

    .. warning:: ``on`` is a so-called "shallow" effect handler that only handles
        the first :func:`~chirho.dynamical.ops.simulate`call within its context,
        and its ``callback`` can be triggered at most once.

    .. warning:: some backends may not support interruptions via arbitrary predicates, and may only support
        interruptions that include additional information such as a statically known time at which to activate.

    :param predicate: A function that takes a state and returns a boolean.
    :param callback: A function that takes a dynamics and state and returns a new dynamics and state.
    :return: A context manager that interrupts a simulation when the predicate is true.
    """
    if callback is None:

        def _on(
            callback: Callable[[Dynamics[T], State[T]], Tuple[Dynamics[T], State[T]]],
        ):
            return on(predicate, callback)

        return _on

    from chirho.dynamical.internals.solver import Interruption

    return Interruption(predicate, callback)



from __future__ import annotations

import functools
from typing import Callable, Hashable, Mapping, Optional, Tuple, TypeVar, Union

T = TypeVar("T")

AtomicIntervention = Union[T, Tuple[T, ...], Callable[[T], Union[T, Tuple[T, ...]]]]
CompoundIntervention = Union[Mapping[Hashable, AtomicIntervention[T]], Callable[..., T]]
Intervention = Union[AtomicIntervention[T], CompoundIntervention[T]]


@functools.singledispatch
def intervene(obs, act: Optional[Intervention[T]] = None, **kwargs):
    """
    Intervene on a value in a probabilistic program.

    :func:`intervene` is primarily used internally in :class:`DoMessenger`
    for concisely and extensibly defining the semantics of interventions. This
    function is generically typed and extensible to new types via
    :func:`functools.singledispatch`. When its first argument is a function,
    :func:`intervene` now behaves like the current `observational.do` effect handler.

    :param obs: a value in a probabilistic program.
    :param act: an optional intervention.
    """
    raise NotImplementedError(f"intervene not implemented for type {type(obs)}")



import contextlib
from math import prod
from typing import Callable, List, Mapping, Optional, Tuple, TypeVar

import pyro
import torch
from torch.utils._pytree import (
    SUPPORTED_NODES,
    PyTree,
    TreeSpec,
    _get_node_type,
    tree_flatten,
    tree_unflatten,
)
from typing_extensions import Concatenate, ParamSpec

P = ParamSpec("P")
Q = ParamSpec("Q")
S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")

ParamDict = Mapping[str, torch.Tensor]


def make_flatten_unflatten(
    v: T,
) -> Tuple[Callable[[T], torch.Tensor], Callable[[torch.Tensor], T]]:
    """
    Returns functions to flatten and unflatten an object. Used as a helper
    in :func:`chirho.robust.internals.linearize.conjugate_gradient_solve`

    :param v: some object
    :raises NotImplementedError:
    :return: flatten and unflatten functions
    :rtype: Tuple[Callable[[T], torch.Tensor], Callable[[torch.Tensor], T]]
    """
    flat_v, treespec = torch.utils._pytree.tree_flatten(v)

    def _flatten(unflat_v: T) -> torch.Tensor:
        parts, _ = torch.utils._pytree.tree_flatten(unflat_v)
        return torch.hstack([x.reshape((x.shape[0], -1)) for x in parts])

    def _unflatten(single_flat_v: torch.Tensor) -> T:
        parts = [
            v_flat.reshape(v.shape)
            for v, v_flat in zip(
                flat_v,
                torch.split(
                    single_flat_v,
                    [int(vi.numel() / flat_v[0].shape[0]) for vi in flat_v],
                    dim=1,
                ),
            )
        ]
        return torch.utils._pytree.tree_unflatten(parts, treespec)

    return _flatten, _unflatten


SPyTree = TypeVar("SPyTree", bound=PyTree)
TPyTree = TypeVar("TPyTree", bound=PyTree)
UPyTree = TypeVar("UPyTree", bound=PyTree)


def pytree_generalized_manual_revjvp(
    fn: Callable[[TPyTree], SPyTree], params: TPyTree, batched_vector: UPyTree
) -> SPyTree:
    """
    Computes the jacobian-vector product using backward differentiation for the jacobian, and then manually
    right multiplying the batched vector. This supports pytree structured inputs, outputs, and params.

    :param fn: function to compute the jacobian of
    :param params: parameters to compute the jacobian at
    :param batched_vector: batched vector to right multiply the jacobian by
    :raises ValueError: if params and batched_vector do not have the same tree structure
    :return: jacobian-vector product
    """

    # Assumptions (in terms of elements of the referenced pytrees):
    # 1. params is not batched, and represents just the inputs to the fn that we'll take the jac wrt.
    #    - params.shape == (*param_shape)
    # 2. batched_vector is the batched vector component of the jv product. It's rightside shape matches params.
    #    - batched_vector.shape == (*batch_shape, *param_shape)
    # 3. The output of the function will have some output_shape, which will cause the jacobian to have shape.
    #    - jac.shape == (*output_shape, *param_shape)
    # So the task is to infer these shapes and line everything up correctly. As a general approach, we'll flatten
    #  the inputs and output shapes in order to apply a standard batched matrix multiplication operation.
    # The output will have shape (*batch_shape, *output_shape).

    # The shaping is complicated by fact that we aren't working with tensors, but PyTrees instead, and we want to
    #  perform the same inner product wrt to the tree structure. This mainly shows up in that the jacobian will
    #  return a pytree with a "root" structure matching that of SPyTree (the return of the fn), but at each leaf
    #  of that tree, we have a pytree matching the structure of TPyTree (the params). This is the tree-structured
    #  equivalent the jac shape matching output on the left, and params on the right.

    jac_fn = torch.func.jacrev(fn)
    jac = jac_fn(params)

    flat_params, param_tspec = tree_flatten(params)

    flat_batched_vector, batched_vector_tspec = tree_flatten(batched_vector)

    if param_tspec != batched_vector_tspec:
        # This is also required by pytorch's jvp implementation.
        raise ValueError(
            "params and batched_vector must have the same tree structure. This requirement generalizes"
            " the notion that the batched_vector must be the correct shape to right multiply the "
            "jacobian."
        )

    # In order to map the param shapes together, we need to iterate through the output tree structure and map each
    #  subtree (corresponding to params) onto the params and batched_vector tree structures, which are both structured
    #  according to the parameters.
    def recurse_to_flattened_sub_tspec(
        pytree: PyTree, sub_tspec: TreeSpec, tspec: Optional[TreeSpec] = None
    ):
        # Default to passed treespec, otherwise compute here.
        _, tspec = tree_flatten(pytree) if tspec is None else (None, tspec)

        # If fn returns a tensor straight away, then the subtree will match at the root node. Check for that here.
        if tspec == sub_tspec:
            flattened, _ = tree_flatten(pytree)
            yield flattened
            return

        # Extract child trees in a node-type agnostic way.
        node_type = _get_node_type(pytree)
        flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
        children_pytrees, _ = flatten_fn(pytree)
        children_tspecs = tspec.children_specs

        # Iterate through children and their specs.
        for child_pytree, child_tspec in zip(children_pytrees, children_tspecs):
            # If we've landed on the target subtree...
            if child_tspec == sub_tspec:
                child_flattened, _ = tree_flatten(child_pytree)
                yield child_flattened  # ...yield the flat child for that subtree.
            else:  # otherwise, recurse to the next level.
                yield from recurse_to_flattened_sub_tspec(
                    child_pytree, sub_tspec, tspec=child_tspec
                )

    flat_out: List[PyTree] = []

    # Recurse into the jacobian tree to find the subtree corresponding to the sub-jacobian for each
    #  individual output tensor in that tree.
    for flat_jac_output_subtree in recurse_to_flattened_sub_tspec(
        pytree=jac, sub_tspec=param_tspec
    ):
        flat_sub_out: List[torch.Tensor] = []

        # Then map that subtree (with tree structure matching that of params) onto the params and batched_vector.
        for i, (p, j, v) in enumerate(
            zip(flat_params, flat_jac_output_subtree, flat_batched_vector)
        ):
            # Infer the parameter shapes directly from passed parameters.
            og_param_shape = p.shape
            param_shape = og_param_shape if len(og_param_shape) else (1,)
            param_numel = prod(param_shape)
            og_param_ndim = len(og_param_shape)

            # Infer the batch shape by subtracting off the param shape on the right.
            og_batch_shape = v.shape[:-og_param_ndim] if og_param_ndim else v.shape
            batch_shape = og_batch_shape if len(og_batch_shape) else (1,)
            batch_ndim = len(batch_shape)

            # Infer the output shape by subtracting off the param shape from the jacobian.
            og_output_shape = j.shape[:-og_param_ndim] if og_param_ndim else j.shape
            output_shape = og_output_shape if len(og_output_shape) else (1,)
            output_numel = prod(output_shape)

            # Reshape for matmul and s.t. that the jacobian can be broadcast over the batch dims.
            j_bm = j.reshape(*(1,) * batch_ndim, output_numel, param_numel)
            v_bm = v.reshape(*batch_shape, param_numel, 1)
            jv = j_bm @ v_bm

            # Reshape result back to the original output shape, with support for empty scalar shapes.
            og_res_shape = (*og_batch_shape, *og_output_shape)
            jv = jv.reshape(*og_res_shape) if len(og_res_shape) else jv.squeeze()

            flat_sub_out.append(jv)

        # The inner product is operating over parameters and the parameter subtree that we just iterated over.
        # So stack these and sum.
        flat_out.append(torch.stack(flat_sub_out, dim=0).sum(0))

    # flat_out is now the flattened version of the tree returned by fn, with each contained tensor having the same
    #  batch dimensions (matching the batching of the batched vector).
    # TODO get out_treespec from the jacobian treespec instead, and don't have an extra forward eval of fn.
    #  Jacobian tree has this structure but its leaves have params treespec.
    out = fn(params)
    _, out_treespec = tree_flatten(out)

    return tree_unflatten(flat_out, out_treespec)


def make_functional_call(
    mod: Callable[P, T],
) -> Tuple[ParamDict, Callable[Concatenate[ParamDict, P], T]]:
    """
    Converts a PyTorch module into a functional call for use with
    functions in :class:`torch.func`.

    :param mod: PyTorch module
    :type mod: Callable[P, T]
    :return: parameter dictionary and functional call
    :rtype: Tuple[ParamDict, Callable[Concatenate[ParamDict, P], T]]
    """
    assert isinstance(mod, torch.nn.Module)
    param_dict: ParamDict = dict(mod.named_parameters())

    @torch.func.functionalize
    def mod_func(params: ParamDict, *args: P.args, **kwargs: P.kwargs) -> T:
        with pyro.validation_enabled(False):
            return torch.func.functional_call(mod, dict(params), args, dict(**kwargs))

    return param_dict, mod_func


@contextlib.contextmanager
def reset_rng_state(rng_state: T):
    """
    Helper to temporarily reset the Pyro RNG state.
    """
    try:
        prev_rng_state: T = pyro.util.get_rng_state()
        yield pyro.util.set_rng_state(rng_state)
    finally:
        pyro.util.set_rng_state(prev_rng_state)



import functools
from typing import TypeVar

import pyro
import pyro.distributions as dist
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin

from chirho.counterfactual.handlers.selection import (
    SelectCounterfactual,
    SelectFactual,
    get_factual_indices,
)
from chirho.counterfactual.internals import no_ambiguity, site_is_ambiguous
from chirho.indexed.ops import gather, get_index_plates, indices_of, scatter
from chirho.observational.ops import observe

T = TypeVar("T")


class FactualConditioningMessenger(pyro.poutine.messenger.Messenger):
    """
    Effect handler for handling ambiguity in conditioning, for use with
    counterfactual semantics handlers such as :class:`MultiWorldCounterfactual` .
    """

    def _pyro_post_sample(self, msg: pyro.poutine.messenger.Message) -> None:

        # expand latent values to include all index plates
        if not msg["is_observed"] and not pyro.poutine.util.site_is_subsample(msg):
            assert isinstance(msg["fn"], TorchDistributionMixin)
            rv, value, event_dim = msg["fn"], msg["value"], len(msg["fn"].event_shape)
            assert value is not None and isinstance(value, torch.Tensor)
            index_plates = get_index_plates()

            new_shape = list(value.shape)
            for k in set(indices_of(rv)) - set(indices_of(value, event_dim=event_dim)):
                dim = index_plates[k].dim
                assert dim is not None
                new_shape = [1] * ((event_dim - dim) - len(new_shape)) + new_shape
                new_shape[dim - event_dim] = rv.batch_shape[dim]

            msg["value"] = value.expand(tuple(new_shape))

    def _pyro_observe(self, msg: dict) -> None:
        if "name" not in msg["kwargs"]:
            msg["kwargs"]["name"] = msg["name"]

        if not site_is_ambiguous(msg):
            return

        msg["value"] = self._dispatched_observe(*msg["args"], name=msg["name"])
        msg["done"] = True
        msg["stop"] = True

    @functools.singledispatchmethod
    def _dispatched_observe(self, rv, obs: torch.Tensor, name: str) -> torch.Tensor:
        raise NotImplementedError

    @_dispatched_observe.register(dist.FoldedDistribution)
    @_dispatched_observe.register(dist.Distribution)
    def _observe_dist(
        self, rv: TorchDistributionMixin, obs: torch.Tensor, name: str
    ) -> torch.Tensor:
        with pyro.poutine.infer_config(config_fn=no_ambiguity):
            with SelectFactual():
                fv = pyro.sample(name + "_factual", rv, obs=obs)

            with SelectCounterfactual():
                cv = pyro.sample(name + "_counterfactual", rv)

            event_dim = len(rv.event_shape)
            fw_indices = get_factual_indices()
            new_value: torch.Tensor = scatter(
                fv, fw_indices, result=cv.clone(), event_dim=event_dim
            )
            new_rv = dist.Delta(new_value, event_dim=event_dim).mask(False)
            return pyro.sample(name, new_rv, obs=new_value)

    @_dispatched_observe.register
    def _observe_tfmdist(
        self, rv: dist.TransformedDistribution, value: torch.Tensor, name: str
    ) -> torch.Tensor:
        tfm = (
            rv.transforms[-1]
            if len(rv.transforms) == 1
            else dist.transforms.ComposeTransformModule(rv.transforms)
        )
        noise_dist = rv.base_dist
        noise_event_dim = len(noise_dist.event_shape)
        obs_event_dim = len(rv.event_shape)

        # factual world
        with SelectFactual(), pyro.poutine.infer_config(config_fn=no_ambiguity):
            new_base_dist = dist.Delta(value, event_dim=obs_event_dim).mask(False)
            new_noise_dist = dist.TransformedDistribution(new_base_dist, tfm.inv)
            obs_noise = pyro.sample(
                name + "_noise_likelihood", new_noise_dist, obs=tfm.inv(value)
            )

        # depends on strategy and indices of noise_dist
        fw = get_factual_indices()
        obs_noise = gather(obs_noise, fw, event_dim=noise_event_dim).expand(
            obs_noise.shape
        )
        # obs_noise = pyro.sample(name + "_noise_prior", noise_dist, obs=obs_noise)
        obs_noise = observe(noise_dist, obs_noise, name=name + "_noise_prior")

        # counterfactual world
        with SelectCounterfactual(), pyro.poutine.infer_config(config_fn=no_ambiguity):
            cf_noise_dist = dist.Delta(obs_noise, event_dim=noise_event_dim).mask(False)
            cf_obs_dist = dist.TransformedDistribution(cf_noise_dist, tfm)
            cf_obs_value = pyro.sample(name + "_cf_obs", cf_obs_dist)

        # merge
        new_value = scatter(
            value, fw, result=cf_obs_value.clone(), event_dim=obs_event_dim
        )
        new_fn = dist.Delta(new_value, event_dim=obs_event_dim).mask(False)
        return pyro.sample(name, new_fn, obs=new_value)



from .interruption import (  # noqa: F401
    DynamicInterruption,
    DynamicIntervention,
    StaticBatchObservation,
    StaticInterruption,
    StaticIntervention,
    StaticObservation,
)
from .trajectory import LogTrajectory  # noqa: F401



from __future__ import annotations

import functools
from typing import Callable, Literal, Optional, Protocol, TypedDict, TypeVar, Union

import pyro
import pyro.distributions as dist
import pyro.distributions.constraints as constraints
import torch
from pyro.distributions.torch_distribution import TorchDistributionMixin
from torch.distributions import biject_to

from chirho.indexed.ops import cond

T = TypeVar("T")

Kernel = Callable[[T, T], torch.Tensor]


@functools.singledispatch
def soft_eq(support: constraints.Constraint, v1: T, v2: T, **kwargs) -> torch.Tensor:
    """
    Computes soft equality between two values ``v1`` and ``v2`` given a distribution constraint ``support``.
    Returns a negative value if there is a difference (the larger the difference, the lower the value)
    and tends to a low value as ``v1`` and ``v2`` tend to each other.

    :param support: A distribution constraint.
    :params v1, v2: the values to be compared.
    :param kwargs: Additional keywords arguments passed further; `scale` adjusts the softness of the inequality.
    :return: A tensor of log probabilities capturing the soft equality between ``v1`` and ``v2``,
            depends on the support and scale.
    :raises TypeError: If boolean tensors have different data types.

    Comment: if the support is boolean, setting ``scale = 1e-8`` results in a value close to ``0.0`` if the values
                are equal and a large negative number ``<=1e-8`` otherwise.
    """
    if not isinstance(v1, torch.Tensor) or not isinstance(v2, torch.Tensor):
        raise NotImplementedError("Soft equality is only implemented for tensors.")
    elif support.is_discrete:
        raise NotImplementedError(
            "Soft equality is not implemented for arbitrary discrete distributions."
        )
    elif support is constraints.real:  # base case
        scale = kwargs.get("scale", 0.1)
        return dist.Normal(0.0, scale).log_prob(v1 - v2)
    else:
        tfm = biject_to(support)
        v1_inv = tfm.inv(v1)
        ldj = tfm.log_abs_det_jacobian(v1_inv, v1)
        v2_inv = tfm.inv(v2)
        ldj = ldj + tfm.log_abs_det_jacobian(v2_inv, v2)
        for _ in range(tfm.codomain.event_dim - tfm.domain.event_dim):
            ldj = torch.sum(ldj, dim=-1)
        return soft_eq(tfm.domain, v1_inv, v2_inv, **kwargs) + ldj


@soft_eq.register
def _soft_eq_independent(support: constraints.independent, v1: T, v2: T, **kwargs):
    result = soft_eq(support.base_constraint, v1, v2, **kwargs)
    for _ in range(support.reinterpreted_batch_ndims):
        result = torch.sum(result, dim=-1)
    return result


@soft_eq.register(type(constraints.boolean))
def _soft_eq_boolean(support, v1: torch.Tensor, v2: torch.Tensor, **kwargs):
    assert support is constraints.boolean
    scale = kwargs.get("scale", 0.1)
    return torch.log(cond(scale, 1 - scale, v1 == v2, event_dim=0))


@soft_eq.register
def _soft_eq_integer_interval(
    support: constraints.integer_interval, v1: torch.Tensor, v2: torch.Tensor, **kwargs
):
    scale = kwargs.get("scale", 0.1)
    width = support.upper_bound - support.lower_bound + 1
    return dist.Binomial(total_count=width, probs=scale).log_prob(torch.abs(v1 - v2))


@soft_eq.register(type(constraints.integer))
def _soft_eq_integer(support, v1: torch.Tensor, v2: torch.Tensor, **kwargs):
    scale = kwargs.get("scale", 0.1)
    return dist.Poisson(rate=scale).log_prob(torch.abs(v1 - v2))


@soft_eq.register(type(constraints.positive_integer))
@soft_eq.register(type(constraints.nonnegative_integer))
def _soft_eq_positive_integer(support, v1: T, v2: T, **kwargs):
    return soft_eq(constraints.integer, v1, v2, **kwargs)


@functools.singledispatch
def soft_neq(support: constraints.Constraint, v1: T, v2: T, **kwargs) -> torch.Tensor:
    """
    Computes soft inequality between two values ``v1`` and ``v2`` given a distribution constraint ``support``.
    Tends to a small value near zero as the difference between the value increases, and tends to
    a large negative value as ``v1`` and ``v2`` tend to each other, summing elementwise over tensors.

    :param support: A distribution constraint.
    :params v1, v2: the values to be compared.
    :param kwargs: Additional keywords arguments:
        `scale` to adjust the softness of the inequality.
    :return: A tensor of log probabilities capturing the soft inequality between ``v1`` and ``v2``.
    :raises TypeError: If boolean tensors have different data types.
    :raises NotImplementedError: If arguments are not tensors.
    """
    if not isinstance(v1, torch.Tensor) or not isinstance(v2, torch.Tensor):
        raise NotImplementedError("Soft equality is only implemented for tensors.")
    elif support.is_discrete:  # for discrete pmf, soft_neq = 1 - soft_eq (in log space)
        return torch.log(-torch.expm1(soft_eq(support, v1, v2, **kwargs)))
    elif support is constraints.real:  # base case
        scale = kwargs.get("scale", 0.1)
        return torch.log(2 * dist.Normal(0.0, scale).cdf(torch.abs(v1 - v2)) - 1)
    else:
        tfm = biject_to(support)
        return soft_neq(tfm.domain, tfm.inv(v1), tfm.inv(v2), **kwargs)


@soft_neq.register
def _soft_neq_independent(support: constraints.independent, v1: T, v2: T, **kwargs):
    result = soft_neq(support.base_constraint, v1, v2, **kwargs)
    for _ in range(support.reinterpreted_batch_ndims):
        result = torch.sum(result, dim=-1)
    return result


class _MaskedDelta(Protocol):
    base_dist: pyro.distributions.Delta
    event_dim: int
    _mask: Union[bool, torch.Tensor]


class _DeterministicReparamMessage(TypedDict):
    name: str
    fn: _MaskedDelta
    value: torch.Tensor
    is_observed: Literal[True]


class KernelSoftConditionReparam(pyro.infer.reparam.reparam.Reparam):
    """
    Reparametrizer that allows approximate soft conditioning on a :func:`pyro.deterministic`
    site using a kernel function that compares the observed and computed values,
    as in approximate Bayesian computation methods from classical statistics.

    This may be useful for estimating counterfactuals in Pyro programs
    corresponding to structural causal models with exogenous noise variables.

    The kernel function should return a score corresponding to the
    log-probability of the observed value given the computed value,
    which is then added to the model's unnormalized log-joint probability
    using :func:`pyro.factor`  :

        :math:`\\log p(v' | v) \\approx K(v, v')`

    The score tensor returned by the kernel function must have shape equal
    or broadcastable to the ``batch_shape`` of the site.

    .. note::
        Kernel functions must be positive-definite and symmetric.
        For example, :class:`~RBFKernel` returns a Normal log-probability
        of the distance between the observed and computed values.
    """

    def __init__(self, kernel: Kernel[torch.Tensor]):
        self.kernel = kernel
        super().__init__()

    def apply(
        self, msg: pyro.infer.reparam.reparam.ReparamMessage
    ) -> pyro.infer.reparam.reparam.ReparamResult:
        assert isinstance(msg["fn"], TorchDistributionMixin)
        assert msg["value"] is not None

        name = msg["name"]
        event_dim = msg["fn"].event_dim
        observed_value = msg["value"]
        computed_value = msg["fn"].base_dist.v

        if observed_value is not computed_value:  # fast path for trivial case
            approx_log_prob = self.kernel(computed_value, observed_value)
            pyro.factor(f"{name}_approx_log_prob", approx_log_prob)

        new_fn = pyro.distributions.Delta(observed_value, event_dim=event_dim).mask(
            False
        )
        return {"fn": new_fn, "value": observed_value, "is_observed": True}


class AutoSoftConditioning(pyro.infer.reparam.strategies.Strategy):
    """
    Automatic reparametrization strategy that allows approximate soft conditioning
    on ``pyro.deterministic`` sites in a Pyro model.

    This may be useful for estimating counterfactuals in Pyro programs corresponding
    to structural causal models with exogenous noise variables.

    This strategy uses :class:`~KernelSoftConditionReparam` to approximate
    the log-probability of the observed value given the computed value
    at each :func:`pyro.deterministic` site whose observed value is different
    from its computed value.

    .. note::
        Implementation details are subject to change.
        Currently uses a few pre-defined kernels such as :class:`~SoftEqKernel`
        and :class:`~RBFKernel` which are chosen for each site based on
        the site's ``event_dim`` and ``support``.
    """

    def __init__(self, *, scale: float = 1.0):
        self.scale = scale
        super().__init__()

    @staticmethod
    def site_is_deterministic(msg: dict) -> bool:
        return (
            msg["is_observed"] is True
            and isinstance(msg["fn"], pyro.distributions.MaskedDistribution)
            and isinstance(msg["fn"].base_dist, pyro.distributions.Delta)
        )

    def configure(self, msg: dict) -> Optional[pyro.infer.reparam.reparam.Reparam]:
        assert isinstance(msg["fn"], TorchDistributionMixin)
        if not self.site_is_deterministic(msg) or msg["value"] is msg["fn"].base_dist.v:
            return None

        support = msg["fn"].base_dist.support

        def _soft_eq(v1: torch.Tensor, v2: torch.Tensor) -> torch.Tensor:
            return soft_eq(support, v1, v2, scale=self.scale)

        return KernelSoftConditionReparam(_soft_eq)

        raise NotImplementedError(
            f"Could not reparameterize deterministic site {msg['name']}"
        )
